"""
Handles most if not all of the data preprocessing.
"""

import multiprocessing as mp

import numpy as np
import scipy.interpolate as interp

from . import utilities as util


def arr_to_tup(a):
    """
    Turns an array into a tuple.

    Parameters
    ----------
    a : arr
        The array to be turned into a tuple.

    Returns
    -------
    tuple
        The tuple form of the converted array.
    """
    return tuple(a.reshape(1, -1)[0])


def find_single_fit_range(lorentz_params):
    """
    Finds the fit range for a single Lorentzian.

    Parameters
    ----------
    lorentz_params : arr
        A 1D lorentzian parameter array.

    Returns
    -------
    tuple
        A two item tuple with the 0 index value being the start of the range
        and the 1 index being the end.
    """
    f0 = lorentz_params[1]
    FWHM = lorentz_params[2]
    return (f0 - 4 * FWHM, f0 - 2 * FWHM, f0 + 2 * FWHM, f0 + 4 * FWHM)


def find_full_fit_range(lorentz_params_array):
    """
    Finds a single fit range given an array defining a group of Lorentzians.

    Parameters
    ----------
    lorentz_params_array : arr
        A 2D Lorentzian parameter array.

    Returns
    -------
    tuple
        A combined fit range tuple.
    """
    (f_low_stop_list, f_low_list, f_high_list,
     f_high_stop_list) = find_all_fit_ranges(lorentz_params_array)
    f_low_stop = min(f_low_stop_list)
    f_low = min(f_low_list)
    f_high = max(f_high_list)
    f_high_stop = max(f_high_stop_list)
    return (f_low_stop, f_low, f_high, f_high_stop)


def find_all_fit_ranges(lorentz_params_array):
    """
    Returns a list of fit ranges for any number of Lorentzians.

    Parameters
    ----------
    lorentz_params_array : arr
        A 2D Lorentzian parameter array.

    Returns
    -------
    tuple
        A four element tuple that's analagous to a fit range. But each item
        is a list containing the corresponding item in a fit range for a given
        Lorentzian.
    """
    f_low_stop_list = []
    f_low_list = []
    f_high_list = []
    f_high_stop_list = []
    for i in range(0, lorentz_params_array.shape[0]):
        fit_range = find_single_fit_range(lorentz_params_array[i])
        f_low_stop_list.append(fit_range[0])
        f_low_list.append(fit_range[1])
        f_high_list.append(fit_range[2])
        f_high_stop_list.append(fit_range[3])
    return (f_low_stop_list, f_low_list, f_high_list, f_high_stop_list)


def check_overlap(lorentz_params_1, lorentz_params_2):
    """
    Checks if two Lorentzians overlap.

    Parameters
    ----------
    lorentz_params_1 : arr
    lorentz_params_2 : arr

    Returns
    -------
    bool
        True if the two Lorentzians sufficiently overlap. False if they don't.
    """
    if lorentz_params_1 is None or lorentz_params_2 is None:
        return False
    [low_lorentz, high_lorentz] = sorted(
        [lorentz_params_1, lorentz_params_2], key=lambda l: l[1])
    low_fit_range = find_single_fit_range(low_lorentz)
    high_fit_range = find_single_fit_range(high_lorentz)
    return low_fit_range[2] > high_fit_range[1]


def partition_lorentz_params_array(lorentz_params_array):
    """
    Returns list of Lorentz arrays.

    Parameters
    ----------
    lorentz_params_array : arr

    Returns
    -------
    list
        Each element in the list is a set of Lorentzians that are clustered
        together.
    """
    lorentz_sets = []
    lorentz_parent = []
    for i in range(0, lorentz_params_array.shape[0]):
        lorentz_sets.append({arr_to_tup(lorentz_params_array[i])})
        lorentz_parent.append(i)
    lorentz_count = lorentz_params_array.shape[0]
    for i in range(0, lorentz_count):
        for j in range(i, lorentz_count):
            l1 = lorentz_params_array[i]
            l2 = lorentz_params_array[j]
            if check_overlap(l1, l2):
                lorentz_parent[j] = lorentz_parent[i]
    for j in range(lorentz_count - 1, -1, -1):
        i = lorentz_parent[j]
        if i != j:
            lorentz_sets[i] = lorentz_sets[i] | lorentz_sets[j]
            lorentz_sets.pop(j)
    for i in range(0, len(lorentz_sets)):
        s = sorted(list(lorentz_sets[i]), key=lambda l: l[1])
        lorentz_params = np.empty((0, 4))
        for l in s:
            l = np.array([l])
            lorentz_params = np.append(lorentz_params, l, axis=0)
        lorentz_sets[i] = lorentz_params
    lorentz_sets = sorted(lorentz_sets, key=lambda a: a[0][1])
    return lorentz_sets


def partition_data_2d(
    data_array,
    fit_range_list,
    lorentz_array_2d,
    scale=(
        0,
        1,
        1024)):
    """
    Partitions a complete generated data set and gives a tuple which has all
    the corresponding data for each Lorentzian cluster as well as how many
    Lorentzians are in each of these clusters.

    Parameters
    ----------
    data_array : arr
        An unedited array of generated data.
    fit_range_list : list
        A tuple of lists corresponding to parts of fit ranges. See
        find_all_fit_ranges for more details.
    lorentz_array_2d : arr
        A 2D Lorentzian array.
    scale : tuple, optional
        The scale to normalize the output data to.

    Returns
    -------
        tuple
            Element 0 is a 2D array with all the data for each Lorentzian
            cluster normalized by the scale factor. Element 1 is the number
            of Lorentzians in each cluster of the same row.
    """
    scale_len = scale[2]
    lorentz_count_array = np.empty((0, 1))
    v_array = np.empty((0, scale_len))
    (f, v) = separate_data(data_array)
    ref_scale = scale_1d(f)
    temp_scale = (0, 1, ref_scale[2])
    for i in range(0, len(fit_range_list)):
        fit_range = fit_range_list[i]
        f_low_range = fit_range[1] - fit_range[0]
        f_high_range = fit_range[2] - fit_range[1]
        f_min = np.random.random() * f_low_range + fit_range[0]
        f_max = np.random.random() * f_high_range + fit_range[2]
        v_norm = normalize_1d(v, temp_scale)
        np.putmask(v_norm, f < f_min, v_norm * 0 - 1)
        np.putmask(v_norm, f > f_max, v_norm * 0 - 1)
        v_clipped = v_norm[v_norm >= 0]
        f_clipped = f[f >= f_min]
        f_clipped = f_clipped[f_clipped <= f_max]
        v_norm = normalize_1d(v_clipped, scale)
        v_array = np.append(v_array, np.array([v_norm]), axis=0)
        num_lorentz = count_lorentz(fit_range, lorentz_array_2d)
        lorentz_count_array = np.append(
            lorentz_count_array, np.array([[num_lorentz]]), axis=0)
    return (v_array, lorentz_count_array)


def count_lorentz(fit_range, lorentz_array_2d):
    """
    Counts how many Lorentzians from the 2D array are within the fit range.

    Parameters
    ----------
    fit_range : tuple
        A fit range tuple.
    lorentz_array_2d : arr
        A 2D Lorentzian array.

    Returns
    -------
    int
        How many of the Lorentzians in the 2D array land within the fit range.
    """
    counter = 0
    for i in range(0, lorentz_array_2d.shape[0]):
        f0 = lorentz_array_2d[i][1]
        if f0 > fit_range[1] and f0 < fit_range[2]:
            counter += 1
    return counter


def find_partitioned_fit_ranges(lorentz_params_list):
    """
    Returns a list of fit ranges from the provided Lorentzians.

    Parameters
    ----------
    lorentz_params_list : list

    Returns
    -------
    list
        A list of all the fit ranges after partitioning the input Lorentzians.
    """
    fit_range_list = []
    for a in lorentz_params_list:
        fit_range_list.append(find_full_fit_range(a))
    return fit_range_list


def evaluate_fit_range(predicted, fit_range):
    """
    Determines if the provided fit ranges are violated.
    """
    test1 = (predicted[0] >= fit_range[0])
    test2 = (predicted[0] <= fit_range[1])
    test3 = (predicted[1] >= fit_range[2])
    test4 = (predicted[1] <= fit_range[3])
    return all([test1, test2, test3, test4])


def evaluate_all_fit_ranges(predicted, fit_range_list):
    """
    Given a list of fit ranges, will check all of them to see how much they
    overlap and are violated.
    """
    tests = []
    for i in range(0, len(fit_range_list)):
        tests.append(evaluate_fit_range(predicted[i], fit_range_list[i]))
    return all(tests)


def disect_lorentz_params_array(lorentz_params_array):
    """
    Returns the number of fit ranges and a list of them from the given array of
    Lorentzian parameters.
    """
    if filter_lorentz(lorentz_params_array):
        lorentz_params_list = partition_lorentz_params_array(
            lorentz_params_array)
        fit_range_list = find_partitioned_fit_ranges(lorentz_params_list)
        num_fit_ranges = len(fit_range_list)
        return (num_fit_ranges, fit_range_list)
    else:
        return (0, None)


def scale_1d(x):
    """
    Determines the scale values for the given 1D array of data.
    Scales are structured as (min value, max value, total number of indices).
    """
    return (min(x), max(x), len(x))


def normalize_index(x, input_scale, output_scale):
    """
    Normalizes indices around the provided scale values.
    """
    return np.round(x / input_scale[2] * output_scale[2])


def normalize_data(
        background_params,
        lorentz_params,
        f,
        v,
        scale=(0, 1, 1024)):
    """
    Normalizes a data set.
    """
    old_f_scale = scale_1d(f)
    old_v_scale = scale_1d(v)
    background_params = background_params[0]
    background_params_norm = normalize_lorentz_1d(
        background_params, old_f_scale, old_v_scale, scale, scale)
    lorentz_params_norm = normalize_lorentz_2d(
        lorentz_params, old_f_scale, old_v_scale, scale, scale)
    f_norm = normalize_1d(f, scale)
    v_norm = normalize_1d(v, scale)
    return np.array([background_params_norm]
                    ), lorentz_params_norm, f_norm, v_norm


def normalize_1d(x, scale=(0, 1, 1024)):
    """
    Normalizes a given array of data around the provided scale factor.
    """
    new_min = scale[0]
    new_max = scale[1]
    new_len = scale[2]
    (min_x, max_x, old_size) = scale_1d(x)
    x_norm = (x - min_x) / (max_x - min_x)
    old_baseline = np.linspace(0, 1, old_size)
    new_baseline = np.linspace(0, 1, new_len)
    if len(old_baseline) <= 1:
        old_baseline = np.array([0, 1])
        x_norm = np.array([1, 0])
    x_interp = interp.interp1d(old_baseline, x_norm)
    x_resized = (x_interp(new_baseline) * (new_max - new_min)) + new_min
    return x_resized


def normalize_0d(x, old_scale=(0, 1, 1024), new_scale=(0, 1, 1024)):
    """
    Normalizes a single value around the provided scale factor.
    """
    old_delta = old_scale[1] - old_scale[0]
    new_delta = new_scale[1] - new_scale[0]
    old_min = old_scale[0]
    new_min = new_scale[0]
    return (x - old_min) * (new_delta / old_delta) + new_min


def normalize_lorentz_1d(
    lorentz, old_f_scale, old_v_scale, new_f_scale=(
        0, 1, 1024), new_v_scale=(
            0, 1, 1024)):
    """
    Normalizes single Lorentzian parameters around the provided scale factors.
    """
    A0 = lorentz[0]
    f0 = lorentz[1]
    G0 = lorentz[2]
    A1 = normalize_0d(A0 + old_v_scale[0], old_v_scale, new_v_scale)
    f1 = normalize_0d(f0, old_f_scale, new_f_scale)
    G1 = normalize_0d(G0 + old_f_scale[0], old_f_scale, new_f_scale)
    return np.array([A1, f1, G1, lorentz[3]])


def normalize_lorentz_2d(
    lorentz, old_f_scale, old_v_scale, new_f_scale=(
        0, 1, 1024), new_v_scale=(
            0, 1, 1024)):
    """
    Normalizes an array of many Lorentzian parameters around the provided scale
    factors.
    """
    lorentz_array = np.empty((0, 4))
    for i in range(0, lorentz.shape[0]):
        l0 = lorentz[i]
        l1 = normalize_lorentz_1d(
            l0, old_f_scale, old_v_scale, new_f_scale, new_v_scale)
        lorentz_array = np.append(lorentz_array, np.array([l1]), axis=0)
    return lorentz_array


def filter_lorentz(lorentz_array):
    """
    Gets rid of empty Lorentzian parameter arrays.
    """
    if lorentz_array is None:
        return False
    else:
        return lorentz_array.shape[1] == 4


def separate_data(f_v_array):
    """
    Given a 2D array of frequency and displacement data, will split it down the
    two like a zipper.
    'ziiiiiiiiiipppppppppppp'
    """
    f = f_v_array[:, 0]
    v = f_v_array[:, 1]
    return (f, v)


def merge_data(f, v):
    """
    Given separate frequency and displacement data, will merge them together
    like a zipper.
    'ziiiiiiiiiipppppppppppp'
    """
    return np.transpose(np.vstack([f, v]))


def separate_all_data(data_arrays_list):
    """
    Separates all data from the provided list of data arrays.
    """
    separated_data_list = []
    for i in range(0, len(data_arrays_list)):
        separated_data_list.append(separate_data(data_arrays_list[i]))
    return separated_data_list


def equalize_data(class_labels, class_data):
    """
    Equalizes the provided data with the provided labels.
    """
    a1 = class_data[np.where(class_labels == 1)[0]]
    a2 = class_data[np.where(class_labels == 2)[0]]
    max_len = min(len(a1), len(a2))
    a1 = a1[0:max_len]
    a2 = a2[0:max_len]
    arr = np.concatenate((a1, a2))
    b1 = np.ones((max_len, 1)) * 0
    b2 = np.ones((max_len, 1)) * 1
    brr = np.concatenate((b1, b2))
    combined_arr = np.append(brr, arr, axis=1)
    combined_arr = np.random.permutation(combined_arr)
    labels = np.transpose(combined_arr)[0]
    data = np.delete(combined_arr, 0, axis=1)
    return labels, data


def pre_process_for_counting(block, scale=(0, 1, 1024)):
    """
    Pre-processes a data set so that it's ready for a counting model to be run
    or trained on it.
    """
    lorentz_arrays_list = block[1]
    data_arrays_list = block[2]
    block_size = len(lorentz_arrays_list)
    scale_list = []
    processed_lorentz_arrays_list = []
    processed_data_array = np.empty((0, scale[2]))
    for i in util._progressbar(range(block_size), "Normalizing: ", 40):
        lorentz_array = lorentz_arrays_list[i]
        data_array = data_arrays_list[i]
        (f_unprocessed, v_unprocessed) = separate_data(data_array)
        scale_f, scale_v = scale_1d(f_unprocessed), scale_1d(v_unprocessed)
        scale_list.append((scale_f, scale_v))
        f_processed = normalize_1d(f_unprocessed, scale)
        v_processed = normalize_1d(v_unprocessed, scale)
        if filter_lorentz(lorentz_array):
            l_processed = normalize_lorentz_2d(
                lorentz_array, scale_f, scale_v, scale, scale)
        else:
            l_processed = None
        processed_lorentz_arrays_list.append(l_processed)
        processed_data_array = np.append(
            processed_data_array, np.array([v_processed]), axis=0)
    results = (processed_lorentz_arrays_list, processed_data_array, scale_list)
    count_labels = []
    pro_length = len(processed_lorentz_arrays_list)
    for i in util._progressbar(range(pro_length), "Labeling: ", 40):
        labels = disect_lorentz_params_array(processed_lorentz_arrays_list[i])
        count_labels.append(labels[0])
    count_labels = np.transpose(np.array([count_labels]))
    count_data = results[1]
    return count_labels, count_data


def classify(input):
    i = input[0]
    lorentz_arrays_list = input[1]
    data_arrays_list = input[2]
    lorentz_params = lorentz_arrays_list[i]
    f_v_data = data_arrays_list[i]
    fit_range_list = disect_lorentz_params_array(lorentz_params)[1]
    (v, labels) = partition_data_2d(f_v_data, fit_range_list, lorentz_params)
    return (labels, v)


def pre_process_for_classifying(block, scale=(0, 1, 1024)):
    """
    Pre-processes a data set so that it's ready for a classifying model to be
    run or trained on it.
    """
    print('\nClassifying Data')
    lorentz_arrays_list = block[1]
    data_arrays_list = block[2]
    block_size = len(lorentz_arrays_list)
    cluster_labels = np.empty((0, 1))
    cluster_data = np.empty((0, scale[2]))
    pool = mp.Pool(mp.cpu_count())
    results = pool.map(
        classify, [
            (i, lorentz_arrays_list, data_arrays_list) for i in range(block_size)])
    pool.close()
    for result in results:
        cluster_labels = np.append(cluster_labels, result[0], axis=0)
        cluster_data = np.append(cluster_data, result[1], axis=0)
    return cluster_labels, cluster_data


def pre_process_for_equal_classifying(block, scale=(0, 1, 1024)):
    """
    Pre-processes a data set so that it's ready for an equal classifying model
    to be run or trained on it.
    """
    class_labels, class_data = pre_process_for_classifying(block, scale)
    eq_labels, eq_data = equalize_data(class_labels, class_data)
    return eq_labels, eq_data


def scale_zoom(x, start, end):
    """
    Zooms in on the data from the provided array given the start and end values
    (not indices).
    """
    length = len(x)
    start_index = int(np.round(length * start))
    end_index = int(np.round(length * end))
    if start_index >= end_index:
        if start_index <= 3:
            start_index = 0
            end_index = 3
        else:
            start_index = end_index - 3
    return normalize_1d(x[start_index:end_index])



"""
Backend for fitting (what I hope is) every concievable type of Lorentzian.
"""

import PySimpleGUI as sg
import numpy as np
from scipy.optimize import least_squares

from . import classify_data as cd
from . import generate_lorentz as gl
from . import utilities as util


def estimate_parameters(f, v, n=1, estimate_background=True):
    """
    Given some frequency and displacement data as well as how many Lorentzians
    are in it, will ballpark some approximate parameters.
    """
    max_f = max(f)
    min_f = min(f)
    max_v = max(v)
    min_v = min(v)
    guess_FWHM = np.ones((1, n)) * (max_f - min_f) / n
    guess_f0 = (np.arange(n).reshape((1, n)) + 1) * \
        (max_f - min_f) / (n + 1) + min_f
    guess_A = np.ones((1, n)) * (max_v - min_v) / n
    guess_phase = np.ones((1, n)) * np.pi
    guess_params = np.concatenate(
        [guess_A, guess_f0, guess_FWHM, guess_phase], axis=0)
    if estimate_background:
        background = np.array([[np.mean(v), 0]])
        guess_params = np.append(guess_params, background)
    guess_params = np.transpose(guess_params)
    return guess_params


def in_lorentz_fit(p, f):
    """
    An in phase Lorentzian for fitting.
    p = array([Amplitude, Center Frequency, Full Width at Half Maximum])
    """
    A = p[0]
    f0 = p[1]
    FWHM = p[2]
    return (A / (2 * FWHM)) * (f - f0) / (((f - f0) / FWHM) ** 2 + 1 / 4)


def out_lorentz_fit(p, f):
    """
    An out of phase Lorentzian for fitting.
    p = array([Amplitude, Center Frequency, Full Width at Half Maximum])
    """
    A = p[0]
    f0 = p[1]
    FWHM = p[2]
    return (A / 4) * (1 / (((f - f0) / FWHM) ** 2 + 1 / 4))


def lorentz_fit(p, f):
    """
    An arbitrary Lorentzian for fitting.
    p = array([Amplitude, Center Frequency, Full Width at Half Maximum, Phase])
    """
    p_no_phase = p[:3]
    return np.cos(p[3]) * in_lorentz_fit(p_no_phase, f) + \
        np.sin(p[3]) * out_lorentz_fit(p_no_phase, f)


def lorentz_residual_fit(p, f, z):
    """
    As lorentz_fit, but with z for the residuals.
    """
    return lorentz_fit(p, f) - z


def multi_lorentz_fit(p, f):
    """
    An arbitrary number of Lorentzians for fitting.
    p = array([
        Amplitude 1, Amplitude 2, ...,
        Center Frequency 1, Center Frequency 2, ...,
        Full Width at Half Maximum 1, Full Width at Half Maximum 2, ...,
        Phase 1, Phase 2, ...])
    """
    slope = p[-1]
    offset = p[-2]
    p_no_background = p[:-2]
    n = int(len(p_no_background) / 4)
    result = slope * f + offset
    for i in range(0, n):
        A = p_no_background[i]
        f0 = p_no_background[i + n]
        FWHM = p_no_background[i + 2 * n]
        phase = p_no_background[i + 3 * n]
        p_single = np.array([A, f0, FWHM, phase])
        result = result + lorentz_fit(p_single, f)
    return result


def multi_lorentz_residual_fit(p, f, z, w):
    """
    As multi_lorentz_fit, but with z for the residuals.
    """
    return (multi_lorentz_fit(p, f) - z) * w


def set_n_least_squares(
        f,
        v,
        n=1,
        noise_filter=0,
        delta_f=None,
        method='lm',
        ftol=1e-10,
        gtol=1e-10,
        xtol=1e-10):
    """
    A least squares fit using the specified number of Lorentzians.
    """
    if delta_f is None:
        delta_f = np.inf
    p_estimate = estimate_parameters(f, v, n)
    bounds = ([None] * len(p_estimate), [None] * len(p_estimate))
    min_f = min(f)
    max_f = max(f)
    max_A = max(v) - min(v)
    for i in range(0, n):
        bounds[0][i] = -2 * max_A
        bounds[1][i] = 2 * max_A
        bounds[0][i + n] = min_f - (max_f - min_f)
        bounds[1][i + n] = max_f + (max_f - min_f)
        bounds[0][i + 2 * n] = -1.5 * (max_f - min_f)
        bounds[1][i + 2 * n] = 1.5 * (max_f - min_f)
        bounds[0][i + 3 * n] = -np.inf
        bounds[1][i + 3 * n] = np.inf
    bounds[0][-2] = -np.inf
    bounds[0][-1] = -np.inf
    bounds[1][-2] = np.inf
    bounds[1][-1] = np.inf
    bounds = (np.array(bounds[0]), np.array(bounds[1]))
    if method == 'lm':
        try:
            fit = least_squares(
                multi_lorentz_residual_fit,
                p_estimate,
                ftol=ftol,
                gtol=gtol,
                xtol=xtol,
                args=(f, v, np.ones(len(f))),
                method='lm')
        except BaseException:
            fit = least_squares(
                multi_lorentz_residual_fit,
                p_estimate,
                ftol=ftol,
                gtol=gtol,
                xtol=xtol,
                args=(f, v, np.ones(len(f))),
                bounds=bounds)
    elif method == 'trf':
        try:
            fit = least_squares(
                multi_lorentz_residual_fit,
                p_estimate,
                ftol=ftol,
                gtol=gtol,
                xtol=xtol,
                args=(f, v, np.ones(len(f))),
                method='trf',
                bounds=bounds)
        except BaseException:
            fit = least_squares(
                multi_lorentz_residual_fit,
                p_estimate,
                ftol=ftol,
                gtol=gtol,
                xtol=xtol,
                args=(f, v, np.ones(len(f))),
                bounds=bounds)
    elif method == 'dogbox':
        try:
            fit = least_squares(
                multi_lorentz_residual_fit,
                p_estimate,
                ftol=ftol,
                gtol=gtol,
                xtol=xtol,
                args=(f, v, np.ones(len(f))),
                method='dogbox',
                bounds=bounds)
        except BaseException:
            fit = least_squares(
                multi_lorentz_residual_fit,
                p_estimate,
                ftol=ftol,
                gtol=gtol,
                xtol=xtol,
                args=(f, v, np.ones(len(f))),
                bounds=bounds)
    else:
        fit = least_squares(
            multi_lorentz_residual_fit,
            p_estimate,
            ftol=ftol,
            gtol=gtol,
            xtol=xtol,
            args=(f, v, np.ones(len(f))),
            bounds=bounds)
    return fit, check_bounds(
        fit.x, bounds, noise_filter=noise_filter, delta_f=delta_f)


def free_n_least_squares(
        f,
        v,
        max_n=3,
        noise_filter=0,
        force_fit=False,
        method='lm'):
    """
    A least squares fit with an unspecified number of Lorentzians.
    """
    delta_f = (max(f) - min(f)) / len(f)
    best_fit, initial_keep = set_n_least_squares(
        f, v, n=1, noise_filter=noise_filter, delta_f=delta_f, method=method)
    n = 1
    while n <= max_n:
        n += 1
        new_fit, keep = set_n_least_squares(
            f, v, n, noise_filter=noise_filter, method=method)
        if util.order_difference(best_fit.cost, new_fit.cost) >= 1 and keep:
            best_fit = new_fit
            initial_keep = keep
        else:
            n = max_n + 1
    if initial_keep or force_fit:
        return best_fit
    else:
        return None


def fit_regions(
        f,
        v,
        regions,
        max_n=3,
        noise_filter=0,
        force_fit=False,
        method='lm'):
    """
    Given frequency data, displacement data, and a regions array, will return a
    list of proposed Lorentzian parameters.
    """
    p_list = []
    for i in util._progressbar(range(0, len(regions)), prefix="Fitting: "):
        region_f, region_v = util.extract_region(i, regions, f, v)
        # plt.plot(region_f, region_v)
        # print(min(region_f))
        # print(max(region_f))
        region_f = cd.normalize_1d(region_f, scale=(
            min(region_f), max(region_f), 1024))
        # print(min(region_f))
        # print(max(region_f))
        # print(min(region_v))
        # print(max(region_v))
        region_v = cd.normalize_1d(region_v, scale=(
            min(region_v), max(region_v), 1024))
        # print(min(region_v))
        # print(max(region_v))
        # plt.plot(region_f, region_v)
        fit = free_n_least_squares(
            region_f,
            region_v,
            noise_filter=noise_filter,
            force_fit=force_fit,
            method=method)
        if fit is not None:
            p_list.append(fit.x)
    return p_list


def extract_parameters(p_list, noise_filter=0):
    """
    Given a list of Lorentzian parameters and a noise level, will return a
    complete and usable numpy array of the parameters.
    """
    p_table = np.empty((0, 4))
    for i in range(0, len(p_list)):
        working_p = p_list[i][0:-2]
        working_p = working_p.reshape(4, len(working_p) // 4)
        working_p = np.transpose(working_p)
        for j in range(0, len(working_p)):
            if np.abs(working_p[j][0]) >= noise_filter:
                p_table = np.append(p_table, working_p, axis=0)
    return p_table[p_table[:, 1].argsort()]


def parameters_from_regions(
        f,
        v,
        regions,
        max_n=3,
        noise_filter=0,
        catch_degeneracies=True,
        allowed_delta_ind=10,
        force_fit=False,
        method='lm'):
    """
    Given regions for analysis, frequency, displacement, will return parameters
    for the fit Lorentzians.
    """
    if len(regions) == 0:
        return np.empty((0, 4))
    p_list = fit_regions(
        f,
        v,
        regions,
        max_n=max_n,
        noise_filter=noise_filter,
        force_fit=force_fit,
        method=method)
    if len(p_list) == 0:
        return np.empty((0, 4))
    p_table = extract_parameters(p_list, noise_filter=noise_filter)
    if catch_degeneracies:
        p_table = remove_degeneracies(p_table, f, allowed_delta_ind=10)
    return correct_parameters(p_table)


def correct_parameters(p_table):
    """
    Puts Lorentzians into a somewhat more standard and consistent form.
    """
    parameters = np.empty((0, 4))
    for i in range(0, len(p_table)):
        p = p_table[i]
        if p[2] < 0:
            A = p[0]
            f0 = p[1]
            FWHM = -p[2]
            phase = np.pi - p[3]
            p = np.array([A, f0, FWHM, phase])
        if p[0] < 0:
            A = -p[0]
            f0 = p[1]
            FWHM = p[2]
            phase = p[3] + np.pi
            p = np.array([A, f0, FWHM, phase])
        if p[3] > 2 * np.pi or p[3] < 0:
            A = p[0]
            f0 = p[1]
            FWHM = p[2]
            phase = np.mod(p[3], 2 * np.pi)
            p = np.array([A, f0, FWHM, phase])
        parameters = np.append(parameters, np.array([p]), axis=0)
    return parameters


def regions_from_parameters(f, p, extension=2):
    regions = np.empty((0, 2))
    for i in range(0, len(p)):
        f_min = p[i][1] - extension * p[i][2]
        f_max = p[i][1] + extension * p[i][2]
        ind_min = util.find_nearest_index(f, f_min)
        ind_max = util.find_nearest_index(f, f_max)
        region = np.array([[ind_min, ind_max]])
        regions = np.append(regions, region, axis=0)
    return regions


def check_bounds(p, bounds, noise_filter=0, delta_f=None):
    """
    Determines if the provided parameters violate the provided bounds.
    """
    if delta_f is None:
        delta_f = np.inf
    keep = True
    for i in range(0, len(p)):
        # Check if violates default bounds.
        if p[i] < bounds[0][i] or p[i] > bounds[1][i]:
            keep = False
    for i in range(0, len(p) // 4):
        # Check if A less than noise level.
        if keep:
            if np.abs(p[i]) < noise_filter:
                keep = False
    for i in range(2 * (len(p) // 4), 3 * (len(p) // 4)):
        # Check if FWHM less than distance between frequencies.
        if keep:
            if np.abs(p[i]) < delta_f:
                keep = False
    return keep


def lorentz_to_data(p, f, v=None, expansion=2):
    """
    Input: One row of extracted (final) parameters, f data, and v data if you
        want to approximate offset.
    Output: The f and v data of the fit Lorentzian within twice the FWHM.
    """
    offset = 0
    min_f_val = p[1] - expansion * np.abs(p[2])
    max_f_val = p[1] + expansion * np.abs(p[2])
    min_f_ind = util.find_nearest_index(f, min_f_val)
    max_f_ind = util.find_nearest_index(f, max_f_val)
    if v is not None:
        offset = np.mean(v[min_f_ind:max_f_ind])
    v = lorentz_fit(p, f)
    return_f = f[min_f_ind:max_f_ind]
    return_v = v[min_f_ind:max_f_ind] + offset
    return return_f, return_v


def lorentz_bounds_to_data(p, f, v, expansion=2):
    """
    Input: One row of extracted (final) parameters, f data, and v data.
    Output: The f and v data of the original data within twice the FWHM.
    """
    min_f_val = p[1] - expansion * np.abs(p[2])
    max_f_val = p[1] + expansion * np.abs(p[2])
    min_f_ind = util.find_nearest_index(f, min_f_val)
    max_f_ind = util.find_nearest_index(f, max_f_val)
    return_f = f[min_f_ind:max_f_ind]
    return_v = v[min_f_ind:max_f_ind]
    return return_f, return_v


def full_lorentz_to_data(p, f, expansion=2):
    """
    Input: One array from the original p_list, f data.
    Output: The f and v data of the fit Lorentzian within twice the FWHM.
    """
    n = len(p) // 4
    f0_arr = p[n:2 * n]
    FWHM_arr = p[2 * n:3 * n]
    min_f_val = min(f0_arr) - expansion * np.abs(max(FWHM_arr))
    max_f_val = max(f0_arr) + expansion * np.abs(max(FWHM_arr))
    min_f_ind = util.find_nearest_index(f, min_f_val)
    max_f_ind = util.find_nearest_index(f, max_f_val)
    v = multi_lorentz_fit(p, f)
    return_f = f[min_f_ind:max_f_ind]
    return_v = v[min_f_ind:max_f_ind]
    return return_f, return_v


def full_lorentz_bounds_to_data(p, f, v, expansion=2):
    """
    Input: One array from the original p_list, f data.
    Output: The f and v data of the original data within twice the FWHM.
    """
    n = len(p) // 4
    f0_arr = p[n:2 * n]
    FWHM_arr = p[2 * n:3 * n]
    min_f_val = min(f0_arr) - expansion * np.abs(max(FWHM_arr))
    max_f_val = max(f0_arr) + expansion * np.abs(max(FWHM_arr))
    min_f_ind = util.find_nearest_index(f, min_f_val)
    max_f_ind = util.find_nearest_index(f, max_f_val)
    return_f = f[min_f_ind:max_f_ind]
    return_v = v[min_f_ind:max_f_ind]
    return return_f, return_v


def catch_degeneracies(p1, p2, f, allowed_delta_ind=10):
    """
    Input: Two elements of p_table and f data.
    Output: Whether or not they are fitting to the same Lorentzian. True if
    they are. False if they aren't.
    """
    delta_ind = util.compare_lorentz(p1, p2, f)
    degenerate = delta_ind < allowed_delta_ind
    return degenerate


def remove_degeneracies(p_table, f, allowed_delta_ind=10):
    """
    Removes duplicate Lorentzians.
    """
    degeneracy_table = []
    new_p_tabel = np.empty((0, 4))
    for i in util._progressbar(
            range(0, len(p_table)),
            prefix="Checking for Degeneracies: "):
        degeneracy_table.append([])
        for j in range(0, len(p_table)):
            if catch_degeneracies(
                    p_table[i],
                    p_table[j],
                    f,
                    allowed_delta_ind=allowed_delta_ind):
                degeneracy_table[i].append(j)
    for i in range(0, len(degeneracy_table)):
        if len(degeneracy_table[i]) > 1:
            for k in range(1, len(degeneracy_table[i])):
                if len(degeneracy_table[i]) > 1:
                    j = degeneracy_table[i][k]
                    degeneracy_table[j] = []
    for i in range(0, len(degeneracy_table)):
        if len(degeneracy_table[i]) > 0:
            new_p_tabel = np.append(
                new_p_tabel, np.array([p_table[i]]), axis=0)
    return new_p_tabel[new_p_tabel[:, 1].argsort()]


def parameters_from_selections(
        data_files,
        region_selections,
        allowed_delta_ind=0,
        noise_filter=0,
        force_fit=False,
        method='lm'):
    region_selections = clean_selections(region_selections)
    sg.one_line_progress_meter_cancel(key='Fitting progress so far:')
    all_peaks = []
    counter = 0
    region_selections, params_selections = region_selections[0], region_selections[1]
    do_fitting = True
    for i in util._progressbar(
            range(0, len(data_files)),
            "Fitting: ",
            progress=False):
        all_peaks.append(np.empty((0, 4)))
        for j in range(0, len(region_selections)):
            try:
                if do_fitting:
                    do_fitting = sg.one_line_progress_meter(
                        'Overall Fitting Progress',
                        counter + 1,
                        len(data_files) *
                        len(region_selections),
                        key='Fitting progress so far:')
                f = data_files[i].f
                v = data_files[i].r
                regions = region_selections[j][i]
                if do_fitting:
                    params = parameters_from_regions(
                        f,
                        v,
                        regions,
                        allowed_delta_ind=allowed_delta_ind,
                        catch_degeneracies=False,
                        noise_filter=noise_filter,
                        force_fit=force_fit,
                        method=method)
                else:
                    pnan = np.array([[np.nan, np.nan, np.nan, np.nan]])
                    params = np.array([[np.nan, np.nan, np.nan, np.nan]])
                    for i in range(len(regions) - 1):
                        params = np.append(params, pnan, axis=1)
                if len(params) == 0:
                    params = np.array([[np.nan, np.nan, np.nan, np.nan]])
            except BaseException:
                params = np.array([[np.nan, np.nan, np.nan, np.nan]])
            all_peaks[i] = np.append(all_peaks[i], params, axis=0)
            counter += 1
    if not do_fitting:
        do_fitting = sg.one_line_progress_meter(
                            'Overall Fitting Progress',
                            len(data_files) * len(region_selections) - 1,
                            len(data_files) * len(region_selections),
                            key='Fitting progress so far:')
    sg.one_line_progress_meter_cancel(key='Fitting progress so far:')
    all_peaks = np.array(all_peaks)
    if params_selections is not None:
        all_peaks = util.append_params_3d(all_peaks, params_selections)
    return denan_parameters(all_peaks)


def spider_fit(data_file, params=None, method='lm'):
    if params is None:
        params = data_file.params
    regions = regions_from_parameters(data_file.f, params)
    x_params = parameters_from_regions(
        data_file.f, data_file.x, regions, method=method)
    y_params = parameters_from_regions(
        data_file.f, data_file.y, regions, method=method)
    return x_params, y_params


def data_file_fit(data_file, params=None, resolution=None):
    x_params, y_params = spider_fit(data_file, params)
    f = data_file.f
    if resolution is not None:
        f = np.linspace(min(f), max(f), resolution)
    x = gl.multi_lorentz_2d(f, x_params)
    y = gl.multi_lorentz_2d(f, y_params)
    new_data_file = util.Data_File(x, y, f)
    new_data_file.import_meta(data_file.name)
    new_data_file.T = data_file.T
    new_data_file.set_params(params)
    return new_data_file


def data_from_region(data_file, region):
    x = data_file.x
    y = data_file.y
    return x[int(region[0]):int(region[1])], y[int(region[0]):int(region[1])]


def data_from_parameters(data_file, params, index=0):
    regions = regions_from_parameters(data_file.f, params)
    return data_from_region(data_file, regions[index])


def denan_parameters(params, nan_level=0.8):
    nan_counters = []
    for i in range(len(params[0])):
        nan_counters.append(0)
    for i in range(len(params)):
        for j in range(len(params[i])):
            if np.isnan(params[i][j][0]):
                nan_counters[j] += 1
    p = []
    for i in range(len(nan_counters)):
        if nan_counters[i] / len(params) > nan_level:
            nan_counters[i] = False
        else:
            nan_counters[i] = True
    for i in range(len(params)):
        p.append([])
        for j in range(len(params[i])):
            if nan_counters[j]:
                p[i].append(params[i][j])
    return np.array(p)


def clean_selections(selections):
    s = selections[0]
    t = []
    p = selections[1]
    for i in range(len(s)):
        if len(s[i]) > 0:
            t.append(s[i])
    return (t, p)



"""
Makes smaller data sets than generate_data that provide the minimum information
needed for model fitting. Generally use this unless you need everything else
that `generate_data` provides.
"""

import multiprocessing as mp
import os
import shutil
import zipfile

import numpy as np

from . import classify_data as cd
from . import generate_lorentz as gl
from . import utilities as util


# import sys


def make_massive_data_set(number, scale=(0, 1, 1024), noise=True):
    """
    Makes a massive set of data.

    Parameters
    ----------
    number : int
        The number of data sets to generate.
    scale : tuple
        Three element scale tuple for normalization. Minimum value, maximum
        value, and then length of the normalized array.
    noise : bool
        If noise should be included.

    Returns
    -------
    tuple
        A three element tuple. First is a list of a background arrays. Then a
        list of Lorentzians, then a list of data arrays.
    """
    background_arrays_list = []
    data_arrays_list = []
    lorentz_list = []
    for i in util._progressbar(range(number), "Generating Data: ", 40):
        background_params, lorentz_params, f, v = gl.generate_data(noise)
        background_arrays_list.append(background_params)
        lorentz_list.append(lorentz_params)
        data_arrays_list.append(cd.merge_data(f, v))
    return (background_arrays_list, lorentz_list, data_arrays_list)


def associate_lorentz(lorentz_array, index):
    """
    Returns an expanded Lorentzian array with the indices of the Lorentzians
    attached to them. This enables functions to remember which Lorentzian is
    which while sorting or processing them.

    Parameters
    ----------
    lorentz_array : arr
        A 2D Lorentzian parameter array.
    index : int
        Index to associate the Lorentzian with.

    Returns
    -------
    arr
        The associated Lorentzian array.
    """
    associated_lorentz_array = np.empty((0, 5))
    for i in range(0, lorentz_array.shape[0]):
        associated_lorentz = np.insert(lorentz_array[i], 0, index)
        associated_lorentz_array = np.append(
            associated_lorentz_array, np.array([associated_lorentz]), axis=0)
    return associated_lorentz_array


def add_set(input):
    """
    Add more data and Lorentzians as defined from the given input.

    Parameters
    ----------
    input : list
        Index, noise, scale.

    Returns
    -------
    list
        New data, new Lorentzians.
    """
    i = input[0]
    noise = input[1]
    scale = input[2]
    background_params, lorentz_params, f, v = gl.generate_data(noise)
    old_f_scale = cd.scale_1d(f)
    old_v_scale = cd.scale_1d(v)
    v_norm = cd.normalize_1d(v, scale)
    l_norm = cd.normalize_lorentz_2d(
        lorentz_params, old_f_scale, old_v_scale, scale, scale)
    l_associated = associate_lorentz(l_norm, i)
    new_data = np.array([v_norm])
    new_lorentz = l_associated
    return [new_data, new_lorentz]


def make_simple_data_set(
        number,
        scale=(0, 1, 1024),
        noise=True,
        progress=True):
    """
    Makes a pre-normalized data set for training off of.

    Parameters
    ----------
    number : int
    scale : tuple, optional
        Three element scale tuple for normalization. Minimum value, maximum
        value, and then length of the normalized array.
    noise : bool, optional
    progress : bool, optional, deprecated

    Returns
    -------
    tuple
        A simple data set in the form of a two element tuple. First are the
        Lorentzians and second is the data.
    """
    print('Generating Data')
    all_data = np.empty((0, scale[2]))
    all_lorentz = np.empty((0, 5))  # Associated Data, A, f0, FWHM, Phase
    pool = mp.Pool(mp.cpu_count())
    results = pool.map(add_set, [(i, noise, scale) for i in range(0, number)])
    pool.close()
    for result in results:
        all_data = np.append(all_data, result[0], axis=0)
        all_lorentz = np.append(all_lorentz, result[1], axis=0)
    return (all_lorentz, all_data)


def make_blank_data_set(number, scale=(0, 1, 1024), noise=True, progress=True):
    """
    Makes a data set without specifying the generated Lorentzians parameters.

    Paremeters
    ----------
    number : int
    scale : tuple, optional
        Three element scale tuple for normalization. Minimum value, maximum
        value, and then length of the normalized array.
    noise : bool, optional
    progress : bool, optional

    Returns
    -------
    tuple
        A simple data set in the form of a two element tuple. There are no
        Lorentzians so the first element is None and the second contains all
        the data.
    """
    all_data = np.empty((0, scale[2]))
    for i in util._progressbar(
            range(number),
            "Generating Data: ",
            40,
            progress=progress):
        background_params, lorentz_params, f, v = gl.generate_data(noise)
        v_norm = cd.normalize_1d(v, scale)
        all_data = np.append(all_data, np.array([v_norm]), axis=0)
    return (None, all_data)


def export_simple_data_set(
        simple_data_set,
        location=os.getcwd(),
        name='simple_set'):
    """
    Exports a generated simple data set to the specified location.

    Parameters
    ----------
    simple_data_set : tuple
    location : string, optional
        The file path to the location. Defaults to current working directory.
    name : string, optional
        Name for the data set with no file extension. Defaults to simple_set.
    """
    export_dir = os.path.join(location, name)
    if os.path.exists(export_dir):
        shutil.rmtree(export_dir)
    os.makedirs(export_dir)
    lorentz_path = os.path.join(export_dir, 'lorentz.csv')
    data_path = os.path.join(export_dir, 'data.csv')
    np.savetxt(lorentz_path, simple_data_set[0], delimiter=',')
    np.savetxt(data_path, simple_data_set[1], delimiter=',')
    shutil.make_archive(export_dir, 'zip', export_dir)
    if os.path.exists(export_dir):
        shutil.rmtree(export_dir)


def import_simple_data_set(path):
    """
    Imports a generated simple data set from the specified location.

    Parameters
    ----------
    path : string

    Returns
    -------
    tuple
        A simple data set in the form of a two element tuple imported from the
        specified file path.
    """
    open_dir = os.path.join(os.getcwd(), 'temp_zip_dir')
    if not os.path.exists(open_dir):
        os.makedirs(open_dir)
    lorentz_path = os.path.join(open_dir, 'lorentz.csv')
    data_path = os.path.join(open_dir, 'data.csv')
    with zipfile.ZipFile(path, "r") as zip_ref:
        zip_ref.extractall(open_dir)
    all_lorentz = np.genfromtxt(lorentz_path, delimiter=',')
    all_data = np.genfromtxt(data_path, delimiter=',')
    shutil.rmtree(open_dir)
    return (all_lorentz, all_data)


def convert_simple_data_set(simp):
    """
    Converts a simple data set to a full one for more involved processing.

    Parameters
    ----------
    simp : tuple
        A simple data set in the form of a two element tuple.

    Returns
    -------
    tuple
        A full data set in the standard form of a list of the backgrounds, then
        a list of the Lorentzians, and finally a list of the full data sets.
    """
    background_arrays_list = None
    data_arrays_list = []
    lorentz_arrays_list = []
    for i in util._progressbar(
        range(
            0,
            simp[1].shape[0]),
        "Converting Displacements: ",
            40):
        data_arrays_list.append(cd.merge_data(
            np.linspace(0, 1, 1024), simp[1][i]))
        lorentz_arrays_list.append(np.empty((0, 4)))
    for i in util._progressbar(
        range(
            0,
            simp[0].shape[0]),
        "Converting Lorentzians: ",
            40):
        l_full = simp[0][i]
        association = int(l_full[0])
        l = l_full[1:5]
        l_arr = lorentz_arrays_list[association]
        l_arr = np.append(l_arr, np.array([l]), axis=0)
        lorentz_arrays_list[association] = l_arr
    return (background_arrays_list, lorentz_arrays_list, data_arrays_list)


def make_single_data_set(
        number,
        scale=(0, 1, 1024),
        noise=True,
        min_noise_amp=1,
        max_noise_amp=1,
        min_noise_width=1,
        max_noise_width=1,
        expansion=2,
        wiggle=0,
        progress=True,
        force_lorentz=False,
        no_norm=False):
    """
    Makes a simple data set of only single Lorentzians.

    Parameters
    ----------
    number : int
    scale : tuple, optional
        Three element scale tuple for normalization. Minimum value, maximum
        value, and then length of the normalized array.
    noise : bool, optional
    min_noise_amp : float, optional
        Minimum amplitude for generated noise.
    max_noise_amp : float, optional
        Maximum amplitude for generated noise.
    min_noise_width : float, optional
        Minimum width for generated noise.
    max_noise_width : float, optional
        Maximum width for generated noise.
    expansion : float, optional
        Number of full widths at half maximum outside that of the Lorentzian to
        include in the data.
    wiggle : float, optional
        How much to offset the Lorentzian from the center of the data.
    progress : bool, optional
        If the progress bar should be shown.
    force_lorentz : bool, optional
        Determines if there must be a Lorentzian generated or not.
    no_norm : bool, optional
        Determines if the data should be normalized or not.

    Returns
    -------
    tuple
        A simple data set in the form of a two element tuple.
    """
    all_data = np.empty((0, scale[2]))
    all_lorentz = np.empty((0, 1))
    F = np.linspace(scale[0], scale[1], scale[2])
    for i in util._progressbar(
        range(number),
        "Generating Data: ",
        40,
            progress=progress):
        has_lorentz = 1
        v, params = gl.generate_lorentz(F)
        if noise:
            noise_amp = np.random.uniform(min_noise_amp, max_noise_amp)
            noise_width = np.random.uniform(min_noise_width, max_noise_width)
            v += gl.generate_noise(F, amount=noise_amp, width=noise_width)
        else:
            v += gl.generate_noise(F, amount=0)
        f = np.linspace(scale[0], scale[1], scale[2])
        old_f_scale = cd.scale_1d(f)
        old_v_scale = cd.scale_1d(v)
        left_wiggle = (np.random.random() - 0.5) * wiggle
        right_wiggle = (np.random.random() - 0.5) * wiggle
        left_expansion = np.random.random() * wiggle + expansion
        right_expansion = np.random.random() * wiggle + expansion
        np.putmask(f, F < params[0][1] -
                   (left_expansion * params[0][2]), f * 0 - 1)
        np.putmask(f, F > params[0][1] +
                   (right_expansion * params[0][2]), f * 0 - 1)
        v = v[f > 0]
        f = f[f > 0]
        if force_lorentz:
            no_lorentz_threshold = 1
        else:
            no_lorentz_threshold = .5
        if np.random.random() > no_lorentz_threshold:
            v -= gl.multi_lorentz_2d(f, params)
            has_lorentz = 0
            if not noise:
                v = np.linspace(scale[0], scale[1], scale[2]) * \
                    (np.random.random() - 0.5) * 2
        v_norm = cd.normalize_1d(v, scale)
        if noise and has_lorentz == 1:
            v_norm += np.linspace(scale[0], scale[1],
                                  scale[2]) * (np.random.random() - 0.5) * 2
        v_norm += ((np.linspace(scale[0], scale[1], scale[2]) + (
            np.random.random() - 0.5) * 30) ** 2) * (np.random.random() - 0.5) * .5
        if not no_norm:
            v_norm = cd.normalize_1d(v_norm, scale)
        all_data = np.append(all_data, np.array([v_norm]), axis=0)
        all_lorentz = np.append(all_lorentz, np.array([[has_lorentz]]), axis=0)
    return (all_lorentz, all_data)


def make_split_data_set(
        number,
        scale=(0, 1, 1024),
        noise=True,
        min_noise_amp=1,
        max_noise_amp=1,
        min_noise_width=1,
        max_noise_width=1,
        expansion=2,
        wiggle=0,
        progress=True):
    """
    Make data set with half of the values forced to have Lorentzians and the
    other not.

    Parameters
    ----------
    number : int
    scale : tuple, optional
        Three element scale tuple for normalization. Minimum value, maximum
        value, and then length of the normalized array.
    noise : bool, optional
    min_noise_amp : float, optional
        Minimum amplitude for generated noise.
    max_noise_amp : float, optional
        Maximum amplitude for generated noise.
    min_noise_width : float, optional
        Minimum width for generated noise.
    max_noise_width : float, optional
        Maximum width for generated noise.
    expansion : float, optional
        Number of full widths at half maximum outside that of the Lorentzian to
        include in the data.
    wiggle : float, optional
        How much to offset the Lorentzian from the center of the data.
    progress : bool, optional
        If the progress bar should be shown.


    Returns
    -------
    tuple
        A simple data set in the form of a two element tuple.
    """
    initial_results = make_single_data_set(
        number,
        scale=scale,
        noise=noise,
        min_noise_amp=min_noise_amp,
        max_noise_amp=max_noise_amp,
        min_noise_width=min_noise_width,
        max_noise_width=max_noise_width,
        expansion=expansion,
        wiggle=wiggle,
        progress=progress,
        no_norm=True)
    additional_results = make_single_data_set(
        number,
        scale=scale,
        noise=noise,
        min_noise_amp=min_noise_amp,
        max_noise_amp=max_noise_amp,
        min_noise_width=min_noise_width,
        max_noise_width=max_noise_width,
        expansion=expansion,
        wiggle=wiggle,
        progress=progress,
        force_lorentz=True,
        no_norm=True)
    all_lorentz = initial_results[0]
    additional_data = additional_results[1]
    initial_data = initial_results[1]
    all_data = np.empty((0, scale[2]))
    for i in range(0, len(all_lorentz)):
        v = cd.normalize_1d(initial_data[i] + additional_data[i], scale)
        all_data = np.append(all_data, [v], axis=0)
    return (all_lorentz, all_data)



"""
Tools to remove the background from real data. All are very experimental.
"""

import PySimpleGUI as sg
import numpy as np
from lorentzian_models import models

from . import fit_lorentz as fl
from . import generate_lorentz as gl
from . import sliding_window as sw
from . import utilities as util


def remove_background(data_files, start_index=0):
    """
    Experimental function to remove background noise and Lorentzians from
    data with smaller peaks in it.

    Parameters
    ----------
    data_files : list
    start_index : int, optional

    Returns
    -------
    f_list : list
    v_list : list
    """
    model = models.tight_lorentzian()
    f_list = []
    v_list = []
    # start_file = data_files[start_index]
    # f = start_file.f
    # v = start_file.r
    # coef_1 = np.polyfit(f, v, 2)
    # bg_1 = np.poly1d(coef_1)
    # v = v - bg_1(f)
    # start_params = auto.quick_analyze(f, v, learn=False, show=True)
    # coef_2 = np.polyfit(f, v - gl.multi_lorentz_2d(f, start_params), 2)
    # bg_2 = np.poly1d(coef_2)
    # v = v - bg_2(f)
    # FWHM = start_params[:,2]
    # min_FWHM = min(FWHM)
    # max_FWHM = max(FWHM)
    # delta_f = max(f) - min(f)
    # min_zoom = int(np.floor(np.log2(delta_f / max_FWHM))) - 1
    min_zoom = 1
    # max_zoom = int(np.ceil(np.log2(delta_f / min_FWHM)))
    max_zoom = 5
    for i in util._progressbar(
            range(0, len(data_files)),
            'Removing Background: '):
        sg.one_line_progress_meter(
            'Background Removeal Progress', i, len(data_files), '-key-')
        # try:
        #     live.close_window()
        # except:
        #     pass
        f = data_files[i].f
        v = data_files[i].r
        # live = lf.Live_Instance(f, v)
        # live.activate()
        coef_1 = np.polyfit(f, v, 2)
        bg_1 = np.poly1d(coef_1)
        v = v - bg_1(f)
        # live = lf.Live_Instance(f, v)
        # live.activate()
        regions = sw.slide_scale(
            model, v, min_zoom=min_zoom, max_zoom=max_zoom)
        while len(regions) > 0:
            bg_params = fl.parameters_from_regions(f, v, regions)
            if len(bg_params) > 0:
                region_deltas = []
                for i in range(0, len(regions)):
                    max_ind = int(regions[i][1])
                    min_ind = int(regions[i][0])
                    max_f = f[max_ind]
                    min_f = f[min_ind]
                    region_deltas.append(max_f - min_f)
                max_delta = max(region_deltas)
                bg_params = np.array(
                    [bg_params[bg_params[:, 2].argsort()][-1]])
                if util.order_difference(bg_params[0][2], max_delta) < 1:
                    # live = lf.Live_Instance(f, v)
                    # live.import_lorentzians(bg_params)
                    # live.activate()
                    if len(bg_params) > 0:
                        v_to_remove = gl.multi_lorentz_2d(f, bg_params)
                        # v_to_remove = cd.normalize_1d(v_to_remove, (min(v), max(v), len(v)))
                        v = v - v_to_remove
                        coef_2 = np.polyfit(f, v, 3)
                        bg_2 = np.poly1d(coef_2)
                        v = v - bg_2(f)
                    regions = sw.slide_scale(
                        model, v, min_zoom=min_zoom, max_zoom=max_zoom)
                else:
                    regions = np.empty((0, 2))
            else:
                regions = np.empty((0, 2))
        # live = lf.Live_Instance(f, v)
        # live.activate()
        # if preview:
        #     live.close_window()
        #     live = lf.Live_Instance(f, v)
        #     live.activate(loop=False)
        f_list.append(f)
        v_list.append(v)
    sg.one_line_progress_meter_cancel('-key-')
    return f_list, v_list



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()
    
setuptools.setup(
    name = 'peak_finder',
    version = '0.5.2',
    author = 'Gabriel Perko-Engel',
    author_email = 'gpe@berkeley.edu',
    description = 'For finding and processing Lorentzian line shapes.',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    license = 'MIT',
    url = 'https://github.com/GabePoel/ML-Peak-Tracker',
    packages = setuptools.find_packages(),
    classifiers = [
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Physics'
    ],
    python_requires = '>=3/6',
    install_requires = [
        'numpy', 
        'scipy',
        'matplotlib', 
        'nptdms', 
        'pysimplegui',
        'tqdm',
    ],
)



"""
Generates complete data sets with all the Lorentzians, background, parameters,
and the like listed. This is generally overkill for model training. Instead,
you should use `efficient_data_generation`.
"""

import os

import numpy as np

from . import generate_lorentz as gl


def mag_round(x, base=10):
    """
    Rounds a number to the nearest order of magnitude.

    Parameters
    ----------
    x : float
        The number to be rounded.
    base : int
        The base that the number should be rounded with respect to.

    Returns
    -------
    float
        The number rounded to the nearest order of magnitude.
    """
    val = 1
    while x > val:
        val *= base
    return val


def get_append(index, count):
    """
    Gives a string to use to the end of large data sets with a bunch of files.
    """
    str_length = len(str(mag_round(count))) - 1
    base_str = str_length * '0'
    num_str = str(base_str[0:str_length - len(str(index))]) + str(index)
    return num_str


def generate_data_set(
        location=os.getcwd(),
        name='generated_data',
        count=1000,
        include_noise=True):
    """
    Generates (by default) 1000 sample data sets in new directory.
    """
    export_dir = os.path.join(location, name)
    if not os.path.exists(export_dir):
        os.makedirs(export_dir)
    for i in range(0, count):
        num_str = get_append(i, count)
        background_params, lorentz_params, f, displacement = gl.generate_data(
            include_noise=include_noise)
        data = np.transpose(np.vstack([f, displacement]))
        background_name = os.path.join(export_dir, num_str + '_background.csv')
        lorentz_name = os.path.join(export_dir, num_str + '_lorentz.csv')
        data_name = os.path.join(export_dir, num_str + '_data.csv')
        np.savetxt(background_name, background_params, delimiter=',')
        np.savetxt(lorentz_name, lorentz_params, delimiter=',')
        np.savetxt(data_name, data, delimiter=',')


def load_data_set(directory=os.path.join(os.getcwd(), 'generated_data')):
    """
    Loads all data from the specified directory.
    Returns as tuple of lists of background arrays, lorentz arrays, and data
    arrays.
    """
    file_names = sorted(os.listdir(directory))
    background_paths = []
    lorentz_paths = []
    data_paths = []
    for i in range(0, len(file_names)):
        name_partition = file_names[i].split('_', 1)
        if name_partition[1] == 'background.csv':
            background_paths.append(file_names[i])
        elif name_partition[1] == 'lorentz.csv':
            lorentz_paths.append(file_names[i])
        elif name_partition[1] == 'data.csv':
            data_paths.append(file_names[i])
    count = len(data_paths)
    background_arrays = []
    lorentz_arrays = []
    data_arrays = []
    for i in range(0, count):
        background_file = os.path.join(directory, background_paths[i])
        lorentz_file = os.path.join(directory, lorentz_paths[i])
        data_file = os.path.join(directory, data_paths[i])
        background_arrays.append(np.genfromtxt(background_file, delimiter=','))
        lorentz_arrays.append(np.atleast_2d(
            np.genfromtxt(lorentz_file, delimiter=',')))
        data_arrays.append(np.genfromtxt(data_file, delimiter=','))
    return (background_arrays, lorentz_arrays, data_arrays)



from .. import train_model as tm
import tensorflow as tf
import os
import sys
root = os.getcwd()
parent = os.path.join(root, '..')
sys.path.insert(1, parent)

# SYLVIA DON'T RUN THIS ONE, I'M NOT USING IT ATM
# Trains several models.
# This model says whether an input is a Lorentzian or not.
# It is only trained on Lorentzians exactly centered within the data
# tightly framed around them.

default_location = os.path.join(parent, 'models')
model_1 = tf.keras.Sequential([
    tf.keras.layers.Dense(1024, activation='relu'),
    tf.keras.layers.Dense(1024, activation='relu'),
    tf.keras.layers.Dropout(0.2),
    tf.keras.layers.Dense(2, activation='softmax')
])

model_2 = tf.keras.Sequential([
    tf.keras.layers.Dense(1024, activation='relu'),
    tf.keras.layers.Dense(1024, activation='relu'),
    tf.keras.layers.Dropout(0.2),
    tf.keras.layers.Dense(2, activation='softmax')
])

model_3 = tf.keras.Sequential([
    tf.keras.layers.Dense(1024, activation='relu'),
    tf.keras.layers.Dense(1024, activation='relu'),
    tf.keras.layers.Dropout(0.2),
    tf.keras.layers.Dense(2, activation='softmax')
])

name = ['simple_class', 'tight_wiggle', 'wide_wiggle']
location = [default_location, default_location, default_location]
data_size = [1000, 1000, 1000]
scale = [(0, 1, 1024), (0, 1, 1024), (0, 1, 1024)]
expansion = [1.5, 1, 1]
noise = [True, True, True]
epochs = [1, 1, 1]
overwrite = [False, False, False]
model_design = [model_1, model_2, model_3]
optimizer = ['Adadelta', 'Adadelta', 'Adadelta']
loss = [None, None, None]
metrics = [['Accuracy'], ['Accuracy'], ['Accuracy']]
stop_condition = [False, False, False]
steps = [1, 1, 1]
wiggle = [0, 2, 10]
verbose = [1, 1, 1]
min_noise_amp = [1, 0.01, 0.01]
max_noise_amp = [1, 5, 5]
min_noise_width = [1, 1, 1]
max_noise_width = [1, 19, 19]
no_quit = [True, True, True]
progress = [True, True, True]
backup = [True, True, True]

tm.passive_multi_train(
    3,
    name,
    location,
    data_size,
    scale,
    expansion,
    noise,
    epochs,
    overwrite,
    model_design,
    optimizer,
    loss,
    metrics,
    stop_condition,
    steps,
    wiggle,
    verbose,
    min_noise_amp,
    max_noise_amp,
    min_noise_width,
    max_noise_width,
    no_quit,
    progress,
    backup)



"""
Peak Finder

A library for finding, fitting to, and tracking Lorentzian peaks from RUS or
other measurements. Tutorials and examples are available at the
`wiki <https://github.com/GabePoel/ML-Peak-Tracker/wiki>`_.

Documentation
-------------
There is some documentation available at th wiki. But, the most fine grained
information is available as docstrings provided within the code itself.

Both the wiki and docstrings assume that `utilities` is imported as `util` and
that `automatic` is imported as `auto`. Everything else is assumed to be
imported as a two letter initialism. For example, the `live_fitting`
subpackage would be imported as `lf`.

Frontend Subpackages
--------------------
automatic
    Quick and dirty machine learning models with little configurability.
automatic_no_ml
    An alternative version of `automatic` for when TensorFlow isn't updated.
live_fitting
    Interactive tools for fitting over imported RUS data.
sliding_window
    More detailed API for peak detecting machine learning models.
utilities
    Core general purpose tools for loading/saving and manipulating RUS data.

Backend Subpackages
-------------------
background_removal
    Experimental techniques to remove background Lorentzians.
classify_data
    Handling of data pre-processing.
data_spy
    Tools for taking a quick little peak at confusing data.
efficient_data_generation
    Alternative tools to create large amounts of training data quickly.
fit_lorentz
    Backend for custom fitting to different types of Lorentzians.
generate_data
    Scripts to generate complete data sets used for training.
generate_lorentz
    Scripts to generate Lorentzians used to train machine learning models.
models
    Definitions of existing machine learning models.
track_peaks
    Automatic peak tracking given initial parameters. The original method.
train_model
    Scripts to automatically train default machine learning models.
"""


from .. import train_model as tm
import tensorflow as tf
import os
import sys
root = os.getcwd()
parent = os.path.join(root, '..')
sys.path.insert(1, parent)

# SYLVIA DON'T RUN THIS ONE, I'M DOING THAT
# Trains the wiggle_hidden_layer_1024 model.
# This model says whether an input is a Lorentzian or not.
# It is trained on Lorentzians that aren't exactly centered and have a
# little 'wiggle' room around them.

model = tf.keras.Sequential([
    tf.keras.layers.Dense(1024, activation='relu'),
    tf.keras.layers.Dense(1024, activation='relu'),
    tf.keras.layers.Dense(1024, activation='relu'),
    tf.keras.layers.Dropout(0.2),
    tf.keras.layers.Dense(2, activation='softmax')
])

root = os.getcwd()
location = os.path.join(parent, 'models')
tm.passive_train(
    name='tight_wiggle',
    location=location,
    data_size=1000,
    scale=(0, 1, 1024),
    expansion=1,
    noise=True,
    epochs=1,
    model_design=model,
    steps=1,
    wiggle=2,
    min_noise_amp=0.01,
    max_noise_amp=5,
    min_noise_width=1,
    max_noise_width=10,
    no_quit=True,
    verbose=1,
    progress=True)



__all__ = [
    '_bit_invert',
    '_matplotlib_mac_fix',
    '_progressbar',
    '_save_file',
    '_scatter_pts',
    '_simplify_regions',
    'append_params_3d',
    'attach_temps_to_parameters',
    'compare_lorentz',
    'Data_File',
    'delete_parameters_from_f_regions_2d',
    'delete_parameters_from_f_regions_3d',
    'delete_parameters',
    'drop_region',
    'extract_region',
    'find_missing_1d',
    'find_missing_2d',
    'find_missing_3d',
    'find_nearest_index',
    'freq_sort_2d',
    'get_all_params',
    'get_freqs',
    'get_temperatures',
    'get_temps_and_freqs',
    'import_tdms_dir',
    'import_tdms_file',
    'import_tdms_files',
    'install_models',
    'load_dir',
    'load_file',
    'load_files',
    'load_freqs_with_temps',
    'load_freqs',
    'load',
    'match_lengths',
    'normalize_1d',
    'order_difference',
    'param_sort',
    'plot_region',
    'remove_nans',
    'save_freqs_with_temps',
    'save_freqs',
    'save_Tf',
    'save',
    'scale_1d',
    'set_all_params',
    'temp_freq_sort_2d',
]

import sys
import os
import pickle
import tkinter as tk
# from typing import overload
import numpy as np
import matplotlib.pyplot as plt
import scipy.interpolate as interp
from tkinter import filedialog
from nptdms import TdmsFile
from time import sleep
from tqdm import tqdm


def _open_file():
    """
    Tkinter backend file dialog for opening a single file.
    """
    tk.Tk().withdraw()
    return filedialog.askopenfilename()


def _open_files():
    """
    Tkinter backend file dialog for opening multiple files.
    """
    tk.Tk().withdraw()
    return filedialog.askopenfilenames()


def _save_file(filters=[], name=''):
    """
    Tkinter backend file dialog for saving a single file.
    """
    tk.Tk().withdraw()
    return filedialog.asksaveasfilename(
        filetypes=filters, initialfile=name)


def _open_folder():
    """
    Tkinter backend file dialog for opening multiple files.
    """
    tk.Tk().withdraw()
    return filedialog.askdirectory()


if 'linux' in sys.platform:
    import gi
    gi.require_version("Gtk", "3.0")
    from gi.repository import Gtk, Gdk, GLib

    def _quick_buttons(dialog):
        """
        Add cancel buttons to GTK dialogs.
        """
        dialog.add_buttons(
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CANCEL)

    def _open_file():
        """
        GTK backend file dialog for opening a single file.
        """
        result = []

        def _open_file_helper(_None):
            """
            Wrapped file opening file dialog window class.
            """
            dialog = Gtk.FileChooserDialog(action=Gtk.FileChooserAction.OPEN)
            _quick_buttons(dialog)
            dialog.add_button(Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
            dialog.run()
            path = dialog.get_filename()
            dialog.destroy()
            Gtk.main_quit()
            result.append(path)
        Gdk.threads_add_idle(GLib.PRIORITY_DEFAULT, _open_file_helper, None)
        Gtk.main()
        return result[0]

    def _open_files():
        """
        GTK backend file dialog for opening multiple files.
        """
        result = []

        def _open_files_helper(_None):
            """
            Wrapped multiple file opening file dialog window class.
            """
            dialog = Gtk.FileChooserDialog(action=Gtk.FileChooserAction.OPEN)
            dialog.set_select_multiple(True)
            _quick_buttons(dialog)
            dialog.add_button(Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
            dialog.run()
            paths = dialog.get_filenames()
            dialog.destroy()
            Gtk.main_quit()
            result.append(paths)
        Gdk.threads_add_idle(GLib.PRIORITY_DEFAULT, _open_files_helper, None)
        Gtk.main()
        return result[0]

    def _save_file(filters=[], name=''):
        """
        GTK backend file dialog for saving a single file.
        """
        result = []

        def _save_file_helper(filters=[], name=''):
            """
            Wrapped file saving file dialog window class.
            """
            dialog = Gtk.FileChooserDialog(action=Gtk.FileChooserAction.SAVE)
            _quick_buttons(dialog)
            dialog.add_button(Gtk.STOCK_SAVE, Gtk.ResponseType.OK)
            # for f in filters:
            #     filter_text = Gtk.FileFilter()
            #     filter_text.set_name(f[0])
            #     dialog.add_filter(filter_text)
            dialog.run()
            path = os.path.join(dialog.get_current_folder(),
                                dialog.get_current_name())
            dialog.destroy()
            Gtk.main_quit()
            result.append(path)
        Gdk.threads_add_idle(GLib.PRIORITY_DEFAULT, _save_file_helper, None)
        Gtk.main()
        return result[0]

    def _open_folder():
        """
        GTK backend file dialog for opening a folder.
        """
        result = []

        def _open_folder_helper(_None):
            """
            Wrapped folder opening file dialog window class.
            """
            dialog = Gtk.FileChooserDialog(
                action=Gtk.FileChooserAction.SELECT_FOLDER)
            _quick_buttons(dialog)
            dialog.add_button(Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
            dialog.run()
            path = dialog.get_filename()
            dialog.destroy()
            Gtk.main_quit()
            result.append(path)
        Gdk.threads_add_idle(GLib.PRIORITY_DEFAULT, _open_folder_helper, None)
        Gtk.main()
        return result[0]

# Holds utilities that many parts of the peak tracker use.


class Data_File:
    """
    Stores the x, y, f, and v data for an imported tdms file.

    Attributes
    ----------
    x : arr
        1D array of the data's x values.
    y : arr
        1D array of the data's y values.
    f : arr
        1D array of the frequencies data is collected at.
    r : arr
        1D array of data's combined magnitudes. See notes. Identical to `v`.
    v : arr
        1D array of data's combined magnitudes. See notes. Identical to `r`.
    T : float
        1D array of temperatures data is collected at. Defaults to probe
        temperature. If the probe temperature is not available, then defaults
        to cryostat temperature. If the cryostat temperature is also not
        available then the temperature is approximated by interpolating
        between the start and end temperatures imported from the TDMS metadata.
    params : arr
        2D parameters array for Lorentzians fitted to the data file.
    stamp : str
        Metadata imported from corresponding TDMS file.
    date : str
        Date at which TDMS file was created.
    time : str
        Time at which TDMS file was created.
    name : str
        Name of the imported TDMS file.
    probe_temp : arr
        1D array of the temperatures of RUS probe.
    cryo_temp : arr
        1D array of the temperatures of RUS cryostat.
    start_temp : float
        Temperature at start of sweep. Imported from TDMS metadata.
    end_temp : float
        Temperature at end of sweep. Imported from TDMS metadata.

    Notes
    -----
    Here `v` is defined as v = sqrt(x ** 2 + y ** 2). This is identical to `r`.
    """

    def __init__(self, x, y, f):
        self.x = x
        self.y = y
        self.f = f
        self.r = np.sqrt(x ** 2 + y ** 2)
        self.v = self.r
        self.params = None
        self.stamp = None
        self.date = None
        self.time = None
        self.name = None
        self.probe_temp = None
        self.cryo_temp = None
        self.start_temp = None
        self.end_temp = None
        self.T = None

    def _import_probe_temp(self, probe_temp):
        probe_temp = remove_nans(probe_temp)
        min_T = min(probe_temp)
        max_T = max(probe_temp)
        len_T = len(self.f)
        self.probe_temp = normalize_1d(probe_temp, (min_T, max_T, len_T))

    def _import_cryo_temp(self, cryo_temp):
        cryo_temp = remove_nans(cryo_temp)
        min_T = min(cryo_temp)
        max_T = max(cryo_temp)
        len_T = len(self.f)
        self.cryo_temp = normalize_1d(cryo_temp, (min_T, max_T, len_T))

    def _import_meta(self, stamp):
        self.stamp = stamp
        try:
            self.date, self.time, self.start_temp, self.end_temp = stamp.split(
                '_')
            self.name = stamp
        except BaseException:
            self.date, self.time, self.start_temp = stamp.split('_')
            self.end_temp = self.start_temp
            self.name = stamp
        self.start_temp = float(self.start_temp[:-1])
        self.end_temp = float(self.end_temp[:-1])

    def _set_temp(self):
        try:
            probe_high = max(self.probe_temp)
            probe_low = min(self.probe_temp)
            if np.abs(
                    probe_high -
                    1.875) < .03 and np.abs(
                    probe_low -
                    1.875) < .03:
                self.T = self.cryo_temp
            elif any(np.isnan(self.probe_temp)):
                self.T = self.cryo_temp
            else:
                self.T = self.probe_temp
        except BaseException:
            self.T = np.linspace(self.start_temp, self.end_temp, len(self.f))

    def get_temp(self, f0):
        """
        Get temperature corresponding to a particular frequency value. The
        temperature is taken from T, however that is determined.

        Parameters
        ----------
        f0 : float
            Frequency value of temperature.
        """
        index = find_nearest_index(self.f, f0)
        return self.T[index]

    def set_params(self, params):
        """
        Attach Lorentzian parameter array to data file.

        Parameters
        ----------
        params : arr
            2D Lorentzian parameter array.
        """
        self.params = params


def set_all_params(data_files, params):
    """
    Set parameters for many sweeps to all items in a list of data files.

    Parameters
    ----------
    data_files : list
        List of data files. Typically corresponding to a set of continuous
        sweeps imported from the same folder.
    params : arr
        3D parameter array with the same number of elements along axis 0 as
        there are data files in the data_files list.

    Notes
    -----
    Also compatible with 2D parameter arrays as long as a single data file is
    provided instead of a list.
    """
    if type(data_files) is list:
        for i in range(len(data_files)):
            data_files[i].set_params(params[i])
    else:
        data_files.set_params(params)


def get_all_params(data_files):
    """
    Provides a single array for the Lorentzian parameters attached to a list
    of data files.

    Parameters
    ----------
    data_files : list
        List of data files. Typically corresponding to a set of continuous
        sweeps imported from the same folder. Must have parameter attributes
        other than None in order to work properly.
    """
    params = []
    if type(data_files) is list:
        for i in range(len(data_files)):
            params.append(data_files[i].params)
    else:
        params = data_files.params
    try:
        return np.array(params)
    except BaseException:
        return params


def scale_1d(x):
    """
    Determines the scale values for the given 1D array of data.

    Parameters
    ----------
    x : arr
        1D array to get normalization from.

    Returns
    -------
    tuple
        Three element scale tuple corresponding to provided array. The first
        is the minimum value of the array, the second is the maximum value of
        the array, and the last is the total number of indices in the provided
        array.

    Notes
    -----
    Scales are structured as `(min_value, max_value, total_number_of_indices)`.
    """
    return (min(x), max(x), len(x))


def match_lengths(arrs):
    """
    Normalizes a set of arrays around the same scale values.

    Parameters
    ----------
    arrs : list
        List of the arrays to normalize.

    Returns
    -------
    list
        List of the normalized arrays. They are sorted in order of most to
        least elements from the original inputs.

    See Also
    --------
    scale_1d : Method to determine normalization scale.
    """
    return_arrs = arrs.copy()
    arrs.sort(key=lambda a: len(a))
    length = len(arrs[0])
    for i in range(0, len(return_arrs)):
        a = return_arrs[i]
        a = normalize_1d(a, (min(a), max(a), length))
        return_arrs[i] = a
    return return_arrs


def normalize_1d(x, scale=(0, 1, 1024)):
    """
    Normalizes a given array of data around the provided scale factor.

    Parameters
    ----------
    x : arr
        1D array to normalize.
    scale : tuple
        Three element scale tuple.

    Returns
    -------
    arr
        The normalized 1D array.

    Notes
    -----
    Scales are structured as `(min_value, max_value, total_number_of_indices)`.
    """
    new_min = scale[0]
    new_max = scale[1]
    new_len = scale[2]
    (min_x, max_x, old_size) = scale_1d(x)
    x_norm = (x - min_x) / (max_x - min_x)
    old_baseline = np.linspace(0, 1, old_size)
    new_baseline = np.linspace(0, 1, new_len)
    if len(old_baseline) <= 1:
        old_baseline = np.array([0, 1])
        x_norm = np.array([1, 0])
    x_interp = interp.interp1d(old_baseline, x_norm)
    x_resized = (x_interp(new_baseline) * (new_max - new_min)) + new_min
    return x_resized


def remove_nans(arr):
    """
    Removes nans from a Lorentzian parameter array of up to two dimensions.

    Parameters
    ----------
    arr : arr
        A 1D or 2D Lorentzian parameter array that may or may non have nans.

    Returns
    -------
    arr
        A new array with all the nans correctly removed.
    """
    if len(arr.shape) == 1:
        return arr[~np.isnan(arr)]
    elif len(arr.shape) == 2:
        new_arr = np.empty((0, len(arr[0])))
        for i in range(0, len(arr)):
            if not any(np.isnan(arr[i])):
                new_arr = np.append(new_arr, np.array([arr[i]]), axis=0)
        return new_arr


def _progressbar(it, prefix="", size=60, file=sys.stdout, progress=True):
    """
    Use with an iteratore as 'it' to show a progress bar while waiting.

    Parameters
    ----------
    it : iterator
        Iterator to build the bar from.
    prefix : str
        What to say before the progress.
    size : int
        Length of bar.
    file : readout
        Where to print the results to.
    progress : bool
        Whether or not to reveal the bar.
    """
    count = len(it)

    def show(j):
        try:
            x = int(size * j / count)
        except BaseException:
            x = 0
        if progress:
            print("%s[%s%s] %i/%i\r\r" %
                  (prefix, "=" * x, "-" * (size - x), j, count), end='\r')
    show(0)
    for i, item in enumerate(it):
        yield item
        show(i + 1)


def load_file(path=None):
    """
    Provides a file dialog to search for a single file. Only provides the path
    of what's selected.

    Parameters
    ----------
    path : str, optional
        Optional file path to use. Will only return the file path if provided.

    Returns
    -------
    str
        Path of the chosen file.

    See Also
    --------
    load : General purpose dialog that actually imports the selected binary.
    """
    if path is None:
        path = _open_file()
    return path


def load_files(paths=[]):
    """
    Provides a file dialog to search for multiple files. Only provides the path
    of what's selected.

    Parameters
    ----------
    path : str, optional
        Optional file path to use. Will only return the file path if provided.

    Returns
    -------
    str
        Path of the chosen file.

    See Also
    --------
    load : General purpose dialog that actually imports the selected binary.
    """
    if len(paths) == 0:
        paths = _open_files()
    return paths


def load_dir(path=None):
    """
    Provides a file dialog to search for a directory. Only provides the path of
    what's selected.

    Parameters
    ----------
    path : str, optional
        Optional file path to use. Will only return the file path if provided.

    Returns
    -------
    str
        Path of the chosen file.

    See Also
    --------
    load : General purpose dialog that actually imports the selected binary.
    """
    if path is None:
        path = _open_folder()
        print(path)
    return path


def import_tdms_file(path=None, show=False):
    """
    Import a TDMS file as a Data_File object.

    Parameters
    ----------
    path : str, optional
        File path to import from. A file dialog will open if no path is
        provided.
    show : bool, optional
        If True then prints information about the imported TDMS file.

    Notes
    -----
    This is a wrapper for `_import_file` that automatically processes it after
    importing. Calling `_import_file()` directly is deprecated and will
    likely be removed.
    """
    data_file, path = _import_file(path=path, show=show, include_path=True)
    stamp = os.path.basename(path)[:-5]
    data_file._import_meta(stamp)
    data_file._set_temp()
    return data_file


def _import_file(path=None, show=False, include_path=False):
    """
    Import a TDMS file as a Data_File object.

    Parameters
    ----------
    path : str, optional
        File path to import from. A file dialog will open if no path is
        provided.
    show : bool, optional
        If True then prints information about the imported TDMS file.

    Notes
    -----
    This is deprecated and will be removed. Please use `import_tdms_file()`.
    """
    if path is None:
        path = _open_file()
    tdms_file = TdmsFile.read(path)
    channels = tdms_file.groups()[0].channels()
    tdms_f = remove_nans(channels[0][:])
    tdms_x = remove_nans(channels[1][:])
    tdms_y = remove_nans(channels[2][:])
    [tdms_f, tdms_x, tdms_y] = match_lengths([tdms_f, tdms_x, tdms_y])
    tdms_file = Data_File(tdms_x, tdms_y, tdms_f)
    if len(channels) >= 5:
        tdms_probe_temp = remove_nans(channels[3][:])
        tdms_cryo_temp = remove_nans(channels[4][:])
        tdms_file._import_probe_temp(tdms_probe_temp)
        tdms_file._import_cryo_temp(tdms_cryo_temp)
    if show:
        print('Imported file from ' + str(path))
    if include_path:
        return tdms_file, path
    return tdms_file


def plot_region(
        i,
        regions,
        f,
        v,
        color=None,
        show_boundaries=False,
        min_color='g',
        max_color='g'):
    """
    Given a region array, some frequency data, and some displacement data, will
    plot the specified index.

    Parameters
    ----------
    i : int
        Index.
    regions : arr
        Array of two element frequency regions.
    f : arr
        1D frequency array.
    v : arr
        1D amplitude array.
    color : str, optional
        Color to plot.
    show_boundaries : bool
        Whether or not to show the provided region boundaries.
    min_color : str, optional
        Color to use for the minimum boundary markers.
    max_color : str, optional
        Color to use for the maximum boundary markers.
    """
    min_f = int(regions[i][0])
    max_f = int(regions[i][1])
    if color is None:
        plt.plot(f[min_f:max_f], v[min_f:max_f])
    else:
        plt.plot(f[min_f:max_f], v[min_f:max_f], color=color)
    if show_boundaries:
        plt.axvline(x=f[min_f], color=min_color)
        plt.axvline(x=f[max_f], color=max_color)


def extract_region(i, regions, f, v):
    """
    Given a region array, some frequency data, and some displacement data, will
    return the frequency and displacement for the region specified by an index.

    Parameters
    ----------
    i : int
        Index.
    regions : arr
        Array of two element frequency regions.
    f : arr
        1D frequency array.
    v : arr
        1D amplitude array.

    Returns
    -------
    arr
        All the frequency values within the specified region.
    arr
        All the amplitude values within the specified region.
    """
    min_f = int(regions[i][0])
    max_f = int(regions[i][1])
    region_f = f[min_f:max_f]
    region_v = v[min_f:max_f]
    return region_f, region_v


def _bit_invert(b):
    """
    Inverts the provided bits.

    Parameters
    ----------
    b : array_like
        Array of 1s and 0s.

    Returns
    arr
        Array of 0s and 1s.
    """
    return np.abs(b - 1)


def drop_region(regions, min_length=10, max_length=None):
    """
    Given a region array, will return only the regions with more than the
    min_length number of indices.

    Parameters
    ----------
    regions : arr
        Array of two element frequency regions.
    min_length : int, optional
        Minimum number of indices to use for filtering regions.
    max_length : int, optional
        Maximum number of indices to use for filtering regions. Defaults to
        None and if left that way will have no maximum.
    """
    kept_regions = np.empty((0, 2))
    if max_length is None:
        max_length = np.inf
    for i in range(0, len(regions)):
        region = regions[i]
        if region[1] - region[0] >= min_length and region[1] - \
                region[0] <= max_length:
            kept_regions = np.append(kept_regions, np.array([region]), axis=0)
    return kept_regions


def order_difference(val_1, val_2):
    """
    Returns how many orders of magnitude val_1 and val_2 differ by.

    Parameters
    ----------
    val_1 : float
    val_2 : float

    Returns
    -------
    float
    """
    return np.abs(np.log10(val_1) - np.log10(val_2))


def compare_lorentz(l1, l2, f):
    """
    Given two Lorentzian parameter arrays will provide a numerical value of how
    similar the two Lorentzians are to each other. This is a measure.

    Parameters
    ----------
    l1 : arr
        1D Lorentzian parameter array.
    l2 : arr
        1D Lorentzian parameter array.

    Returns
    -------
    float
    """
    f1 = l1[1]
    f2 = l2[1]
    delta_f0 = np.abs(f2 - f1)
    delta_f = max(f) - min(f)
    ind_per_f = len(f) / delta_f
    delta_ind = ind_per_f * delta_f0
    return np.abs(delta_ind)


def find_nearest_index(arr, val):
    """
    Returns the index of the value in an array closest to a provided value.

    Parameters
    ----------
    arr : arr
        1D array to search from.
    val : float
        Value to search for.

    Returns
    -------
    int
    """
    reduced_arr = np.abs(arr - val)
    min_reduced_val = min(reduced_arr)
    return np.where(reduced_arr == min_reduced_val)[0][0]


def _simplify_regions(regions):
    """
    Simplifies set of regions.

    Parameters
    ----------
    regions : arr
        Array of two element frequency regions.

    Returns
    -------
    arr
        The simplified array of two index frequency regions.
    """
    region_line = []
    for i in range(0, len(regions)):
        region_line.append([regions[i][0], 'start', False])
        region_line.append([regions[i][1], 'end', False])

    def take_first(elem):
        return elem[0]
    region_line.sort(key=take_first)
    j = 0
    last_index = None
    while j < len(region_line):
        if j == len(region_line) - 1:
            next_index = None
        else:
            next_index = region_line[j + 1][1]
        this_index = region_line[j][1]
        if last_index is None or next_index is None:
            region_line[j][2] = True
        elif this_index == 'start' and last_index == 'end':
            region_line[j][2] = True
        elif this_index == 'end' and next_index == 'start':
            region_line[j][2] = True
        last_index = this_index
        j += 1
    region_starts = []
    region_ends = []
    for i in range(0, len(region_line)):
        if region_line[i][2]:
            if region_line[i][1] == 'start':
                region_starts.append(region_line[i][0])
            else:
                region_ends.append(region_line[i][0])
    simplified_regions = np.empty((0, 2))
    for i in range(0, len(region_starts)):
        start = region_starts[i]
        end = region_ends[i]
        simplified_regions = np.append(
            simplified_regions, np.array([[start, end]]), axis=0)
    return simplified_regions


def param_sort(params_2d):
    """
    Sorts a 2D parameter array in order of frequency.

    Parameters
    ----------
    params_2d : arr
        2D Lorentzian parameter array.

    Returns
    -------
    arr
        A 2D Lorentzian parameter array sorted by frequencies.
    """
    not_nan_params = np.empty((0, 4))
    nan_params = np.empty((0, 4))
    nan_indices = []
    for i in range(0, len(params_2d)):
        if any(np.isnan(params_2d[i])):
            nan_indices.append(i)
            nan_params = np.append(nan_params, [params_2d[i]], axis=0)
        else:
            not_nan_params = np.append(not_nan_params, [params_2d[i]], axis=0)
    not_nan_params = not_nan_params[not_nan_params[:, 1].argsort()]
    final_params = np.empty((0, 4))
    offset = 0
    for i in range(0, len(params_2d)):
        if i in nan_indices:
            offset += 1
            final_params = np.append(
                final_params, [[np.nan, np.nan, np.nan, np.nan]], axis=0)
        else:
            final_params = np.append(
                final_params, [not_nan_params[i - offset]], axis=0)
    return final_params


def append_params_3d(p1, p2, force=False):
    """
    Adds new Lorentzians to an existing set of parameters. Does not add later
    data points of existing Lorentzians!

    Parameters
    ----------
    p1 : arr
        3D Lorentzian parameter array.
    p2 : arr
        3D Lorentzian parameter array.
    force : bool, optional
        Determines whether or not to force a merged 3D paramter array should be
        created even if the two provided arrays aren't normally compatible.

    Returns
    -------
    arr
        Combined 3D Lorentzian parameter array.
    """
    p3 = []
    if p2 is None or 0 in np.array(p2).shape:
        return p1
    if p1 is None or 0 in np.array(p1).shape:
        return p2
    if not force:
        if not len(p1) == len(p2):
            raise ValueError("Parameters not of the same length.")
        else:
            for i in range(0, len(p1)):
                combined = np.append(p1[i], p2[i], axis=0)
                p3.append(combined)
    else:
        if len(p1) > len(p2):
            p3 = p1
        else:
            p3 = p2
    return np.array(p3)


def save(some_object, path=None, name=''):
    """
    Saves any python object as a pickled binary.

    Parameters
    ----------
    some_object : obj
        An object to pickle and save as a `.pkl` file.
    path : str, optional
        Where to save the binary to. Will open a file dialog if no path is
        provided.
    name : str
        The name of the file to save. Will blank in file dialog if no name is
        provided.

    Returns
    -------
    str
        The path the object has been saved to.

    See Also
    --------
    load : Function to load objects saved in this way.
    """
    if path is None:
        path = _save_file(filters=(("python objects", "*.pkl"),
                                   ("all files", "*.*")), name=name)
    pickle.dump(some_object, open(path, "wb"))
    return path


def load(path=None):
    """
    Loads any pickled python object from binary form.

    Parameters
    ----------
    path : str, optional
        File path of saved binary to load pickled object from.

    Returns
    -------
    obj
        Whatever the saved object is.

    See Also
    --------
    save : Function to save objects this is capable of loading.
    """
    if path is None:
        path = _open_file()
    return pickle.load(open(path, "rb"))


def import_tdms_files(paths=[], show=True):
    """
    Import multiple TDMS files as a list of data files. This is the recommended
    way to import some, but not all, TDMS files within a given directory.

    Parameters
    ----------
    paths : list
        List of strings for the file paths of all the TDMS files that should be
        imported.
    show : bool, optional
        Whether or not to print information about the imported TDMS files.

    Returns
    -------
    list
        List of the data file objects corresponding to the imported TDMS files.

    See Also
    --------
    Data_File : Definition of a data file object.
    """
    paths = load_files(paths)
    data_files = []
    if show:
        print(paths)
    i = 0
    with tqdm(total=len(paths), file=sys.stdout) as pbar:
        for p in paths:
            pbar.set_description('Imported: %d' % (1 + i))
            pbar.update(1)
            sleep(0.01)
            if p[-5:] == '.tdms':
                stamp = os.path.basename(p)[:-5]
                data_file = _import_file(p)
                data_file._import_meta(stamp)
                data_file._set_temp()
                data_files.append(data_file)
    data_files.sort(key=lambda d: int(str(d.date) + str(d.time)))
    print('Imported file order:')
    for i in range(len(data_files)):
        pre = str(i) + ': ' + \
            ((len(str(len(data_files) - 1)) - len(str(i))) * ' ')
        print(pre + str(data_files[i].stamp))
    return data_files


def import_tdms_dir(path=None, show=True):
    """
    Import multiple TDMS files as a list of data files. This is the recommended
    way to import all TDMS files within a given directory.

    Parameters
    ----------
    paths : list
        List of strings for the file paths of all the TDMS files that should be
        imported.
    show : bool, optional
        Whether or not to print information about the imported TDMS files.

    Returns
    -------
    list
        List of the data file objects corresponding to the imported TDMS files.

    See Also
    --------
    Data_File : Definition of a data file object.
    """
    path = load_dir(path)
    names = os.listdir(path)
    data_files = []
    if show:
        print(path)
    i = 0
    with tqdm(total=len(names), file=sys.stdout) as pbar:
        for name in names:
            pbar.set_description('Imported: %d' % (1 + i))
            pbar.update(1)
            sleep(0.01)
            if name[-5:] == '.tdms':
                stamp = name[:-5]
                file_path = os.path.join(path, name)
                data_file = _import_file(file_path)
                data_file._import_meta(stamp)
                data_file._set_temp()
                data_files.append(data_file)

    data_files.sort(key=lambda d: int(str(d.date) + str(d.time)))
    return data_files


def get_temperatures(data_files):
    """
    Get a temperature array from a list of data files.

    Parameters
    ----------
    data_files : list
        List of data files.

    Returns
    -------
    arr
        Array of the starting temperature for the sweep represented by each
        data file. Has the same number of elements as the provided list of
        data files.
    """
    temperatures = []
    for i in range(0, len(data_files)):
        temperatures.append(float(data_files[i].start_temp))
    return np.array(temperatures)


def get_freqs(data_files):
    """
    Get array of frequencies from the Lorentzians fitted to a list of data
    files.

    Parameters
    ----------
    data_files : list
        List of data files.

    Returns
    -------
    arr
        2D array of the frequencies corresponding to the fitted Lorentzians
        attached to the provided list of data files. Axis 0 determines which
        sweep is being referenced and axis 1 has the frequencies for all the
        peaks fitted within a given sweep. When displayed as a table, the
        columns show the movement of a peak across all the sweeps.

    See Also
    --------
    get_temps_and_freqs : As this but with temperatures as well.
    save_freqs : Save peak frequenceis found this way as a CSV.
    save_freqs_with_temps : Save peak frequencies and temperatures as a CSV.
    """
    p = get_all_params(data_files)
    f = []
    for i in range(0, len(data_files)):
        f.append(p[i][..., 1])
    return freq_sort_2d(np.array(f))


def get_temps_and_freqs(data_files):
    """
    Get array of frequencies and temperatures from the Lorentzians fitted to a
    list of data files.

    Parameters
    ----------
    data_files : list
        List of data files.

    Returns
    -------
    arr
        2D array of the frequencies corresponding to the fitted Lorentzians
        attached to the provided list of data files. Axis 0 determines which
        sweep is being referenced and axis 1 has the frequencies for all the
        peaks fitted within a given sweep. When displayed as a table, the
        columns show the movement of a peak across all the sweeps. In this
        format the temperatures as shown in the first column.

    See Also
    --------
    get_freqs : As this but without temperatures. All data is frequencies.
    save_freqs : Save peak frequencies as a CSV.
    save_freqs_with_temps : Save peak frequencies and temperatures as a CSV.
    """
    f = get_freqs(data_files)
    T = np.transpose([get_temperatures(data_files)])
    fT = np.append(T, f, axis=1)
    return fT


def _matplotlib_mac_fix():
    """
    Fix tkinter backend on older macs.
    """
    if 'mac' in sys.platform:
        import matplotlib
        import importlib
        matplotlib.use("TkAgg")
        import matplotlib.pyplot as plt
        importlib.reload(plt)


def _scatter_pts(pts, ref_arr, tar_arr):
    """
    Tool to create array of points for `Point_Selector` to display as an
    interactive scatter plot.

    Parameters
    ----------
    pts : arr
        1D frequency array of points to plot.
    ref_arr : arr
        1D frequency array.
    tar_arr : arr
        1D amplitude array.

    Returns
    -------
    arr
        Amplitude points to preview in scatter plot.

    See Also
    --------
    plot_points : Method of lf._Point_Selector
    """
    arr = np.empty((0,))
    for i in range(len(pts)):
        j = find_nearest_index(pts[i], ref_arr)
        arr = np.append(arr, [tar_arr[j]])
    return arr


def save_freqs(f, name='freqs_kHz', alter=True):
    """
    Save frequencies as a CSV.

    Parameters
    ----------
    f : arr
        1D array of frequencies to be saved.
    name : str, optional
        Name of the saved CSV.
    alter : bool, optional
        If True, saves as kHz. If False, saves as Hz.
    """
    if alter:
        f = np.sort(f) / 1000
    path = _save_file(
        filters=(
            ("text file", "*.txt"),
            ("all files", "*.*")),
        name=name)
    np.savetxt(path, f, delimiter='\n', fmt='%10.15f')


def save_freqs_with_temps(data_files, name='temp_K_and_freqs_kHz'):
    """
    Save frequencies and temperatures as a CSV. The rows correspond to sweeps
    and the colums correspond to individual peaks. The first column is the
    temperature each sweep was taken at. This saves them from the data files
    directly.

    Parameters
    ----------
    data_files : list
        List of data files.
    name : str, optional
        Name of the CSV.
    """
    T = []
    for i in range(len(data_files)):
        T.append(np.mean(data_files[i].T))
    p = get_all_params(data_files)
    f = p[:, :, 1] / 1000
    arr = np.append(np.transpose([T]), f, axis=1)
    arr = temp_freq_sort_2d(arr)
    path = _save_file(
        filters=(
            ("text file", "*.txt"),
            ("comma separated values", "*.csv"),
            ("all files", "*.*")),
        name=name)
    np.savetxt(path, arr, delimiter=',', fmt='%10.15f')


def save_Tf(Tf, path=None, name='temp_K_and_freqs_kHz'):
    """
    Save frequencies and temperatures as a CSV. The rows correspond to sweeps
    and the colums correspond to individual peaks. The first column is the
    temperature each sweep was taken at. This saves them from as array of
    temperatures and frequencies.

    Parameters
    ----------
    Tf : arr
        2D array of temperatures and frequencies.
    path : str, optional
        File path for saved CSV. Loads a file dialog if no path is provided.
    name : str, optional
        Name to use for saved CSV.

    See Also
    --------
    get_temps_and_freqs : Function to get array used for `save_Tf`.
    """
    if path is None:
        path = _save_file(
            filters=(
                ("text file", "*.txt"),
                ("comma separated values", "*.csv"),
                ("all files", "*.*")),
            name=name)
    T = np.transpose(np.array([Tf[:, 0]]))
    f = Tf[:, 1:] / 1000
    Tf = np.append(T, f, axis=1)
    np.savetxt(path, Tf, delimiter=',', fmt='%10.15f')


def load_freqs(path=None):
    """
    Load a 1D array of frequencies. Assumes that they are saved as kHz.

    Parameters
    ----------
    path : str, optional
        File path to load from. Loads a file dialog if no path is provided.

    See Also
    --------
    save_freqs : Function to save frequencies loaded by `load_freqs`.
    """
    if path is None:
        path = _open_file()
    f = np.loadtxt(path, delimiter='\n')
    return f * 1000


def load_freqs_with_temps(path=None):
    """
    Loads temperatures and frequencies as an array.

    Parameters
    ----------
    path : str, optional
        File path to load from. Loads a file dialog if no path is provided.

    Returns
    -------
    arr
        2D array of temperatures and frequencies.

    See Also
    --------
    save_Tf :
        Function to save CSVs usable by `load_freqs_with_temps` starting with
        the same format that it creates.
    save_freqs_with_temps :
        Function to save CSVs usable by `load_freqs_with_temps` starting with
        data files.
    """
    if path is None:
        path = _open_file()
    Tf = np.loadtxt(path, delimiter=',')
    f = np.transpose(Tf)[1:] * 1000
    T = np.array([np.transpose(Tf)[0]])
    Tf = np.append(T, f, axis=0)
    return np.transpose(Tf)


def attach_temps_to_parameters(data_files):
    """
    Makes extended parameter array that includes the temperatures they were
    measured at.

    Parameters
    ----------
    data_files : list
        List of data files.

    Returns
    -------
    arr
        Extended parameter array.
    """
    p = get_all_params(data_files)
    q = []
    for i in range(len(p)):
        q.append([])
        for j in range(len(p[i])):
            p0 = p[i][j]
            f0 = p0[1]
            if np.isnan(f0):
                T0 = np.nan
            else:
                index = find_nearest_index(data_files[i].f, f0)
                T0 = data_files[i].T[index]
            p0 = np.append(p0, [T0])
            q[i].append(p0)
    return np.array(q)


def delete_parameters(params_3d, index):
    """
    Given a 3D Lorentzian parameter array, removes a specific sweep as
    specified by index.

    Parameters
    ----------
    params_3d : arr
        3D Lorentzian parameter array.
    index : int
        Index to remove.

    Returns
    -------
    arr
        3D Lorentzian parameter array with the given sweep removed.
    """
    p = []
    for i in range(len(params_3d)):
        p.append([])
        for j in range(len(params_3d[i])):
            if not j == index:
                p[i].append(params_3d[i][j])
    return np.array(p)


def delete_parameters_from_f_regions_3d(parameters_3d, f_regions):
    """
    Removes the parameters within specific regions from a 3D Lorentzian
    parameter array.

    Parameters
    ----------
    parameters_3d : arr
        3D Lorentzian parameter array to remove Lorentzians from.
    f_regions : arr
        Array of two element frequency regions.

    Returns
    -------
    arr
        3D Lorentzian parameter array with the given Lorentzians removed.
    """
    p = []
    for i in range(len(parameters_3d)):
        p_2d = delete_parameters_from_f_regions_2d(
            parameters_3d[i], f_regions[i])
        p.append(p_2d)
    return np.array(p)


def delete_parameters_from_f_regions_2d(parameters_2d, f_region):
    """
    Removes the parameters within specific regions from a 2D Lorentzian
    parameter array.

    Parameters
    ----------
    parameters_2d : arr
        2D Lorentzian parameter array to remove Lorentzians from.
    f_regions : arr
        Array of two element frequency regions.

    Returns
    -------
    arr
        2D Lorentzian parameter array with the given Lorentzians removed.
    """
    p = []
    for i in range(len(parameters_2d)):
        if parameters_2d[i][1] >= f_region[0] and parameters_2d[i][1] <= f_region[1]:
            p.append([np.nan, np.nan, np.nan, np.nan])
        else:
            p.append(parameters_2d[i])
    return np.array(p)


def freq_sort_2d(f_2d):
    """
    Sort a 2D frequency array. This would likely be an array specifically
    containing peaks as found in `get_freqs`. Axis 0 corresponds to the sweeps
    and axis 1 corresponds to the frequencies within each sweep. This sorts
    the peaks for all sweeps with respect to the mean value of that peak's
    location over all sweeps. So, there may be certain sweeps where the peaks
    appear to be out of order. But, that only happens because several peaks
    cross each other due to a change in temperature, pressure, or under some
    field.

    Parameters
    ----------
    f_2d : arr
        2D frequency array.

    Returns
    -------
    arr
        Sorted 2D frequency array.

    See Also
    --------
    get_freqs : Function to get 2D frequency arrays.
    """
    f_2d = np.transpose(f_2d)
    final_sorting_order = []
    for i in range(len(f_2d)):
        mean_freq = np.mean(remove_nans(f_2d[i]))
        final_sorting_order.append((i, mean_freq))
    # print(np.array(final_sorting_order))
    final_sorting_order.sort(key=lambda t: t[1])
    # print(np.array(final_sorting_order))
    sorted_freqs = []
    for i in range(len(f_2d)):
        sorted_freqs.append(f_2d[final_sorting_order[i][0]])
    sorted_freqs = np.transpose(np.array(sorted_freqs))
    return sorted_freqs


def temp_freq_sort_2d(Tf_2d):
    """
    As `freq_sort_2d` but with compatibility for the first element within each
    sweep being the starting temperature of that sweep.

    Parameters
    ----------
    Tf_2d : arr
        2D temperature and frequency array.

    Returns
    -------
    arr
        Sorted 2D temperature and frequency array.

    See Also
    --------
    freq_sort_2d : Function to sort just frequencies.
    get_temps_and_freqs : Function to get arrays that would be sorted.
    """
    T = np.transpose(Tf_2d)[0]
    f = np.transpose(Tf_2d)[1:]
    f = freq_sort_2d(np.transpose(f))
    Tf = np.append([T], np.transpose(f), axis=0)
    Tf = np.transpose(Tf)
    return Tf


def find_missing_3d(params_3d, overlap=1, degree=2, sanity_check=True):
    """
    Fills in the nans in the input array. For a given section nans, finds a
    polynomial fit over the region both before and after that region. Then
    uses a transformed sinusoidal connection function to smoothly transition
    between those two polynomials and project the values into the nans. This
    works best for input arrays of smoothly changing functions. Nans at the
    ends are not filled in.

    Parameters
    ----------
    params_3d : arr
        A 3d Lorentzian parameter array.
    overlap : int, optional
        Number of indices above and below to consider, by default 1.
    degree : int, optional
        Maximum degree of polynomial to fit, by default 2. Note that the
        degree is lowered depending on the numbed of points available to
        produce the fit so that there is a meaningful unique fit in each case.

    Returns
    -------
    arr
        The array with the nans filled in.
    """
    new_list = []
    for i in range(params_3d.shape[1]):
        p = params_3d[:, i]
        new_list.append(
            find_missing_2d(
                p,
                overlap=overlap,
                degree=degree,
                sanity_check=sanity_check))
    new_misshapen = np.array(new_list)
    new_params = []
    for i in range(params_3d.shape[0]):
        new_params.append(new_misshapen[:, i])
    return np.array(new_params)


def find_missing_2d(params_2d, overlap=1, degree=2, sanity_check=True):
    """
    Fills in the nans in the input array. For a given section nans, finds a
    polynomial fit over the region both before and after that region. Then
    uses a transformed sinusoidal connection function to smoothly transition
    between those two polynomials and project the values into the nans. This
    works best for input arrays of smoothly changing functions. Nans at the
    ends are not filled in.

    Parameters
    ----------
    params_2d : arr
        A 2d Lorentzian parameter array.
    overlap : int, optional
        Number of indices above and below to consider, by default 1.
    degree : int, optional
        Maximum degree of polynomial to fit, by default 2. Note that the
        degree is lowered depending on the numbed of points available to
        produce the fit so that there is a meaningful unique fit in each case.

    Returns
    -------
    arr
        The array with the nans filled in.
    """
    A = find_missing_1d(params_2d[..., 0], overlap=overlap, degree=degree,
                        sanity_check=sanity_check)
    f0 = find_missing_1d(params_2d[..., 1], overlap=overlap, degree=degree,
                         sanity_check=sanity_check)
    FWHM = find_missing_1d(params_2d[..., 2], overlap=overlap, degree=degree,
                           sanity_check=sanity_check)
    phase = find_missing_1d(params_2d[..., 3], overlap=overlap, degree=degree,
                            sanity_check=sanity_check)
    p = np.transpose(np.array([A, f0, FWHM, phase]))
    return p


def find_missing_1d(params_1d, overlap=1, degree=2, sanity_check=True):
    """
    Fills in the nans in the input array. For a given section nans, finds a
    polynomial fit over the region both before and after that region. Then
    uses a transformed sinusoidal connection function to smoothly transition
    between those two polynomials and project the values into the nans. This
    works best for input arrays of smoothly changing functions. Nans at the
    ends are not filled in.

    Parameters
    ----------
    params_1d : arr
        A 1d array.
    overlap : int, optional
        Number of indices above and below to consider, by default 1.
    degree : int, optional
        Maximum degree of polynomial to fit, by default 2. Note that the
        degree is lowered depending on the numbed of points available to
        produce the fit so that there is a meaningful unique fit in each case.

    Returns
    -------
    arr
        The array with the nans filled in.
    """
    cuts = _make_cuts(params_1d)
    mod_params = list(params_1d)
    fillings = []
    for cut in cuts:
        start = _search_down(params_1d, cut[0] - 1, overlap)
        end = _search_up(params_1d, cut[-1] + 1, overlap)
        filling = [start, end]
        fillings.append(filling)
    for i in range(len(cuts)):
        c = cuts[i]
        f = fillings[i]
        start_check = np.abs(c[0] - f[0]) > 1
        end_check = np.abs(c[-1] - f[-1]) > 1
        if start_check and end_check:
            original_piece = params_1d[f[0]:f[-1] + 1]
            param_piece = _filling_function(original_piece, degree=degree)
            allow = not any(param_piece > max(original_piece))
            allow = allow and not any(param_piece < min(original_piece))
            if degree == 0 or (sanity_check and (not allow)):
                p0 = params_1d[c[0] - 1]
                p1 = params_1d[c[-1] + 1]
                param_piece = np.linspace(p0, p1, f[-1] - f[0] + 1)
        elif not (c[0] == f[0] or c[-1] == f[-1]):
            p0 = params_1d[c[0] - 1]
            p1 = params_1d[c[-1] + 1]
            param_piece = np.linspace(p0, p1, f[-1] - f[0] + 1)
        else:
            param_piece = np.full((f[-1] - f[0] + 1), np.nan)
        for j in range(len(param_piece)):
            j_original = j + f[0]
            if not np.isnan(
                    param_piece[j]) and np.isnan(
                    mod_params[j_original]):
                mod_params[j_original] = param_piece[j]
    return np.array(mod_params)


def _make_cuts(params_1d):
    """
    Makes a list of lists specifying the nans in the given array. If the a
    given sublist has two elements then they specify the first and last indices
    that are nans within the corresponding sub-array of nans. If the sublist
    has only one index then the corresponding sub-array is only one element
    long but that element is still a nan.

    Parameters
    ----------
    params_1d : arr
        Array to return the nan cuts of.

    Returns
    -------
    list
        The list of lists.
    """
    indices = np.arange(len(params_1d))
    indices_to_replace = indices[np.isnan(params_1d)]
    try:
        cuts = [[indices_to_replace[0]]]
        for i in range(len(indices_to_replace) - 1):
            if indices_to_replace[i + 1] - indices_to_replace[i] == 1:
                cuts[-1].append(indices_to_replace[i + 1])
            else:
                cuts.append([indices_to_replace[i + 1]])
        if len(cuts[-1]) == 0:
            cuts = cuts[:-1]
    except:
        cuts = []
    return cuts


def _search_up(arr, start, count):
    """
    Searches upwards from the start index until it hits a nan. Returns the
    highest non-nan index within the specified search distance.

    Parameters
    ----------
    arr : arr
        Array to search.
    start : int
        Index to start search from.
    count : int
        Maximum distance to search.

    Returns
    -------
    int
        Found index.
    """
    top = len(arr) - 1
    a = arr[start + 1:]
    step = 0
    actual_steps = 0
    while step < count:
        step += 1
        if len(a) > 0 and not np.isnan(a[0]):
            actual_steps += 1
            a = a[1:]
    return min(top, start + actual_steps)


def _search_down(arr, start, count):
    """
    Searches downwards from the start index until it hits a nan. Returns the
    lowest non-nan index within the specified search distance.

    Parameters
    ----------
    arr : arr
        Array to search.
    start : int
        Index to start search from.
    count : int
        Maximum distance to search.

    Returns
    -------
    int
        Found index.
    """
    a = arr[:start]
    step = 0
    actual_steps = 0
    while step < count:
        step += 1
        if len(a) > 0 and not np.isnan(a[-1]):
            actual_steps += 1
            a = a[:-1]
    return max(0, start - actual_steps)


def _filling_connection_function(x):
    """
    Helper function for `_filling_weight_function` to smoothly connect from 0
    to 1.

    Parameters
    ----------
    x : arr
        Array of x values.

    Returns
    -------
    arr
        Array of C(x) values for connection function C(x).

    See Also
    --------
    _filling_function : Function that this feeds into.
    """
    return np.sin(np.pi * (x - 1 / 2)) / 2 + 1 / 2


def _filling_weight_function(x, x1, x2):
    """
    Function that smoothly fills from 0 to 1 between x1 and x2.

    Parameters
    ----------
    x : arr
        Array of x values.
    x1 : float
        Starting value to fill between.
    x2 : float
        Ending value to fill between.
    """
    out = []
    for x0 in x:
        if x0 <= x1:
            out.append(0)
        elif x0 >= x2:
            out.append(1)
        else:
            out.append(_filling_connection_function((x0 - x1) / (x2 - x1)))
    return np.array(out)


def _filling_combination_function(x, x1, x2, f1, f2):
    """
    Function that smoothly connects polynomials f1 and f2 from x1 to x2.

    Parameters
    ----------
    x : arr
        Domain.
    x1 : float
        Starting value.
    x2 : float
        Ending value.
    f1 : arr
        Coefficients for polynomial representation of starting function.
    f2 : [type]
        Coefficients for polynomial representation of ending funciton.

    Returns
    -------
    arr
        Output with functions connected.
    """
    w = _filling_weight_function

    def g1(x):
        x_outs = []
        for x0 in x:
            x_out = 0
            for i in range(len(f1)):
                x_out += f1[i] * x0 ** (len(f1) - i - 1)
            x_outs.append(x_out)
        return x_outs

    def g2(x):
        x_outs = []
        for x0 in x:
            x_out = 0
            for i in range(len(f2)):
                x_out += f2[i] * x0 ** (len(f2) - i - 1)
            x_outs.append(x_out)
        return x_outs
    return w(x, x1, x2) * g1(x) + (1 - w(x, x1, x2)) * g2(x)


def _filling_function(y, degree=2):
    """
    Fit over the nan gap.

    Parameters
    ----------
    y : arr
        All values of array including nans to be filled between.
    degree : int, optional
        Degree of polynomial function to fit, by default 1. The degree will
        automatically be reduced if there are not enough points provided to
        make a meaningful unique fit.

    Returns
    -------
    arr
        The original y input but with all the nans filled in.
    """
    if np.isnan(y[0]) or np.isnan(y[-1]):
        return y
    else:
        i = 0
        while not np.isnan(y[i]):
            i += 1
        y_start = y[:i]
        j = len(y) - 1
        while not np.isnan(y[j - 1]):
            j -= 1
        y_end = y[j:]
        d1 = min(degree, len(y_start) - 1)
        d2 = min(degree, len(y_end) - 1)
        x_start = np.arange(len(y_start))
        x_end = np.arange(len(y_end)) + j
        fit_start = np.polyfit(x_start, y_start, d1)
        fit_end = np.polyfit(x_end, y_end, d2)
        x_mid = np.arange(j - i) + i
        y_mid = _filling_combination_function(
            x_mid, i - 1, j, fit_start, fit_end)
        return np.concatenate([y_start, y_mid, y_end])


def install_models():
    """
    Install pre-made Lorentzian machine learning models.
    """
    command_string = 'pip3 install git+https://github.com/GabePoel/'
    'Lorentzian-Models#egg=lorentzian_models --user'
    os.system(command_string)


"""
A set of tools that let you "spy" into the data you're working with to see how
the models are working.
"""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.gridspec import GridSpec

from . import classify_data as cd
from . import fit_lorentz as fl
from . import utilities as util


# from matplotlib.widgets import Button


def plot_count(i, count_predictions, count_label, count_data):
    """
    Visualizes Lorentzian counting predictions.
    """
    true_label = int(count_label[i][0])
    predictions_array = count_predictions[i]
    v = count_data[i]
    plt.grid(False)
    plt.xticks([])
    plt.yticks([])
    f = np.linspace(0, 1, 1024)
    plt.plot(f, v)
    predicted_label = np.argmax(predictions_array)
    if predicted_label == true_label:
        color = 'green'
    elif np.abs(predicted_label - true_label) == 1:
        color = 'blue'
    else:
        color = 'red'
    plt.xlabel(
        "{} {:2.0f}% ({})".format(
            str(predicted_label) +
            ' lorentzian cluster:',
            100 *
            np.max(predictions_array),
            str(true_label)),
        color=color)


def plot_value_array(i, count_predictions, count_label):
    """
    Shows the predicted value for a Lorentzian counting prediction.
    """
    predictions_array = count_predictions[i]
    true_label = int(count_label[i][0])
    plt.grid(False)
    plt.xticks([])
    plt.yticks([0, .1, .2, .3, .4, .5, .6, .7, .8, .9, 1])
    thisplot = plt.bar(range(len(predictions_array)), predictions_array)
    plt.ylim([0, 1])
    predicted_label = np.argmax(predictions_array)
    thisplot[predicted_label].set_color('red')
    thisplot[true_label].set_color('blue')


def plot_count_result(i, count_predictions, count_label, count_data):
    """
    Plots the results after counting Lorentzians within given regions.
    """
    plt.figure(figsize=(20, 6))
    plt.subplot(1, 2, 1)
    plot_count(i, count_predictions, count_label, count_data)
    plt.subplot(1, 2, 2)
    plot_value_array(i, count_predictions, count_label)


def preview_data(i, data_set):
    """
    Plots and prints some basic stats about a generated data set.
    """
    lorentz_params = data_set[1][i]
    data = data_set[2][i]
    (f, v) = cd.separate_data(data)
    f = cd.normalize_1d(f, (0, 1, 1024))
    v = cd.normalize_1d(v, (0, 1, 1024))
    plt.plot(f, v)
    print(cd.disect_lorentz_params_array(lorentz_params)[0])


def scatter_data_files(data_files):
    for i in range(0, len(data_files)):
        scale = (0, 1, len(data_files[i].r))
        f = data_files[i].f
        r = cd.normalize_1d(data_files[i].r, scale)
        just_plotted = plt.plot(f, r + i, alpha=0.5)
        color = just_plotted[0].get_color()
        if not data_files[i].params is None:
            pts_f = data_files[i].params[..., 1]
            pts_r = util.scatter_pts(pts_f, f, r)
            plt.scatter(pts_f, pts_r + i, color=color)
        plt.text(f[0], r[0] + i, str(data_files[i].T[0]) +
                 ' K ', color=color, ha='right')
        plt.text(f[-1], r[-1] + i, ' ' + str(data_files[i].T[-1]) +
                 ' K', color=color, ha='left')


def spider_plot(data_file, params=None, bg=True, color_1=None, color_2=None):
    if bg:
        plt.plot(data_file.x, data_file.y, color=color_1)
    if params is None:
        params = data_file.params
    if params is not None:
        regions = fl.regions_from_parameters(data_file.f, params)
        for region in regions:
            plt.plot(data_file.x[int(region[0]):int(region[1])], data_file.y[int(
                region[0]):int(region[1])], color=color_2)


def plot_all_params(params, index):
    gs = GridSpec(3, 2)
    fig = plt.figure()
    ax0 = fig.add_subplot(gs[0, 0])
    ax1 = fig.add_subplot(gs[0, 1])
    ax2 = fig.add_subplot(gs[1, 0])
    ax3 = fig.add_subplot(gs[1, 1])
    ax4 = fig.add_subplot(gs[2, :])
    ax0.set_title('Amplitude')
    ax1.set_title('Position')
    ax2.set_title('Full Width at Half Maximum')
    ax3.set_title('Phase')
    ax0.plot(params[:, index, 0], picker=True)
    ax1.plot(params[:, index, 1], picker=True)
    ax2.plot(params[:, index, 2], picker=True)
    ax3.plot(params[:, index, 3], picker=True)
    ax4.plot(np.linspace(0, 1, 100))

    def on_pick(event):
        print(event.artist.get_xdata())
        print(event.artist.get_ydata())

    fig.canvas.mpl_connect('pick_event', on_pick)



from .. import train_model as tm
import tensorflow as tf
import os
import sys
root = os.getcwd()
parent = os.path.join(root, '..')
sys.path.insert(1, parent)

# SYLVIA DON'T RUN THIS ONE, I'M NOT USING IT ATM
# Trains the mid_hidden_layer_1024 model.
# This model says whether an input is a Lorentzian or not.
# It is only trained on Lorentzians exactly centered within the data
# tightly framed around them.

model = tf.keras.Sequential([
    tf.keras.layers.Dense(1024, activation='relu'),
    tf.keras.layers.Dense(1024, activation='relu'),
    tf.keras.layers.Dropout(0.2),
    tf.keras.layers.Dense(2, activation='softmax')
])

location = os.path.join(parent, 'models')
tm.passive_train(
    name='simple_class',
    location=location,
    data_size=5000,
    scale=(0, 1, 1024),
    expansion=1.5,
    noise=True,
    epochs=50,
    model_design=model,
    steps=1,
    no_quit=True)



"""
Tools for tracking peaks in later sweeps based on peaks in earlier sweeps.
"""

import PySimpleGUI as sg
import numpy as np
from lorentzian_models import models

from . import automatic
from . import fit_lorentz as fl
from . import live_fitting as lf
from . import sliding_window as sw
from . import utilities as util


def find_last_peak(params_3d, index):
    """
    Finds the parameters for the last time a given Lorentzian wasn't a nan
    array.

    Parameters
    ----------
    params_3d : arr
        3D Lorentzian parameter array.
    index : int
        Index corresponding to the 2D Lorentzian parameter array to extract the
        last peak from.

    Returns
    -------
    arr
        1D Lorentzian parameter array for the last peak from the chosen index.
    arr
        Array with the number of recursion steps.
    """
    # print(params_3d)
    if not any(np.isnan(params_3d[-1][index])):
        # This is gross. It should be fixed at some point and made cleaner.
        return_params_1d = params_3d[-1][index]
        count_1d = np.array([1])
        if len(params_3d) > 1:
            comp_params, comp_count = find_last_peak(params_3d[0:-1], index)
            if (return_params_1d == comp_params).all():
                return_params_1d, count_1d = find_last_peak(
                    params_3d[0:-1], index)
    else:
        return_params_1d, count_1d = find_last_peak(params_3d[0:-1], index)
    return return_params_1d, count_1d + 1


def get_reference_params(params_3d):
    """
    Finds the last usable parameters for a complete 2d Lorentzian parameter
    array.

    Parameters
    ----------
    params_3d : arr
        3D Lorentzian parameter array.

    Returns
    -------
    arr
        2D Lorentzian parameter array.
    arr
        Step count.
    """
    reference_params_2d = np.empty((0, 4))
    step_count = np.empty((0, 1))
    for i in range(0, len(params_3d[0])):
        params_1d, count_1d = find_last_peak(params_3d, i)
        reference_params_2d = np.append(
            reference_params_2d, np.array([params_1d]), axis=0)
        step_count = np.append(step_count, [count_1d], axis=0)
    return reference_params_2d, step_count


def find_min_f(f, active_params, active_index):
    """
    Finds lowest f that doesn't intersect another Lorentzian.

    Parameters
    ----------
    f : arr
        1D array of frequencies.
    active_params : arr
        Lorentzian parameter array corresponding to the given frequency array.
    active_index : int
        Recursion index.

    Returns
    -------
    float
        Minimum frequency.
    """
    if active_index == 0:
        return min(f)
    elif not np.isnan(active_params[active_index - 1][1]):
        return active_params[active_index - 1][1] + \
            active_params[active_index - 1][2]
    else:
        return find_min_f(f, active_params, active_index - 1)


def find_max_f(f, active_params, active_index):
    """
    Finds greatest f that doesn't intersect another Lorentzian.

    Parameters
    ----------
    f : arr
        1D array of frequencies.
    active_params : arr
        Lorentzian parameter array corresponding to the given frequency array.
    active_index : int
        Recursion index.

    Returns
    -------
    float
        Minimum frequency.
    """
    if active_index == len(active_params) - 1:
        return max(f)
    elif not np.isnan(active_params[active_index + 1][1]):
        return active_params[active_index + 1][1] - \
            active_params[active_index + 1][2]
    else:
        return find_max_f(f, active_params, active_index + 1)


def find_extremes_f(f, active_params, active_index):
    """
    Finds values for an f region that doesn't intersect other Lorentzians.

    Parameters
    ----------
    f : arr
        1D array of frequencies.
    active_params : arr
        Lorentzian parameter array corresponding to the given frequency array.
    active_index : int
        Recursion index.

    Returns
    -------
    float
        Minimum frequency.
    """
    f1 = find_min_f(f, active_params, active_index)
    f2 = find_max_f(f, active_params, active_index)
    min_f = min(f1, f2)
    max_f = max(f1, f2)
    return min_f, max_f


def recursive_split(model, f, v, regions, base_zoom, limit=None):
    """
    Split regions recursively.

    Parameters
    ----------
    model : model
        Model to use.
    f : arr
        1D frequency array.
    v : arr
        1D amplitude (voltage) array.
    regions : arr
        Regions array.
    base_zoom : float
        Zoom level to recursively split over.
    limit : float
        The zoom limit level that ends the recursive splitting.

    Returns
    -------
    arr
        Recursively split regions.
    """
    if limit is None:
        limit = base_zoom - 2
    if base_zoom > 0:
        try:
            potential_regions = sw.split_peaks(
                model, f, v, regions, base_zoom - 1, base_zoom + 1)
        except BaseException:
            if base_zoom >= limit:
                potential_regions = recursive_split(
                    model, f, v, regions, base_zoom - 1, limit=limit)
            else:
                potential_regions = np.empty((0, 2))
    else:
        potential_regions = np.empty((0, 2))
    return potential_regions


def find_missing_peaks_sep(
        f,
        v,
        active_params,
        reference_params,
        search_limit):
    """
    Uses machine learning to search within a given frequency space for missing
    Lorentzians.

    Parameters
    ----------
    f : arr
        1D frequency array.
    v : arr
        1D amplitude (voltage) array.
    active_params : arr
        Lorentzian parameter array.
    reference_params : arr
        Lorentzian parameter array.
    search_limit : float
        Multiplicative factor saying how far out to search.
    """
    updated_params = np.empty((0, 4))
    noise_level = 3 * sw.extract_noise(f)
    model = models.tight_lorentzian()
    for i in range(0, len(active_params)):
        if not any(np.isnan(active_params[i])):
            updated_params = np.append(
                updated_params, np.array([active_params[i]]), axis=0)
        else:
            min_f, max_f = find_extremes_f(f, active_params, i)
            lim_f1 = reference_params[i][1] - \
                search_limit * reference_params[i][2]
            lim_f2 = reference_params[i][1] + \
                search_limit * reference_params[i][2]
            min_lim_f = min(lim_f1, lim_f2)
            max_lim_f = max(lim_f1, lim_f2)
            min_f = max(min_f, min_lim_f)
            max_f = min(max_f, max_lim_f)
            min_f, max_f = min(min_f, max_f), max(min_f, max_f)
            min_ind = util.find_nearest_index(f, min_f)
            max_ind = util.find_nearest_index(f, max_f)
            reference_FWHM = reference_params[i][2]
            delta_f = max_f - min_f
            try:
                base_zoom = np.abs(
                    int(np.round(np.log2(delta_f / reference_FWHM))))
            except BaseException:
                print(reference_params[i])
                print(delta_f)
                print(min_f)
                print(max_f)
                base_zoom = 8
            # make sure we don't waste computing resources looking at nothing
            if base_zoom > 8:
                base_zoom = 8
            if base_zoom < 2:
                base_zoom = 2
            regions = np.array([[min_ind, max_ind]])
            potential_regions = recursive_split(
                model, f, v, regions, base_zoom)
            if len(potential_regions) > 0:
                potential_params = fl.parameters_from_regions(
                    f, v, potential_regions, noise_filter=noise_level)
                if len(potential_params) > 0:
                    f0_vals = potential_params[:, 1]
                    reference_f0 = reference_params[i][1]
                    closest_index = util.find_nearest_index(
                        f0_vals, reference_f0)
                    closest_params = potential_params[closest_index]
                    updated_params = np.append(
                        updated_params, np.array([closest_params]), axis=0)
                else:
                    updated_params = np.append(updated_params, np.array(
                        [[np.nan, np.nan, np.nan, np.nan]]), axis=0)
            else:
                updated_params = np.append(updated_params, np.array(
                    [[np.nan, np.nan, np.nan, np.nan]]), axis=0)
    return updated_params


def find_missing_peaks_tog(
        f,
        v,
        active_params,
        reference_params,
        search_limit):
    if not any(np.isnan(active_params.flatten())):
        return active_params
    else:
        noise_level = 3 * sw.extract_noise(f)
        model = models.tight_lorentzian()
        nan_indices = []
        nan_limits = []
        regions = np.empty((0, 2))
        known_zooms = []
        for i in range(0, len(active_params)):
            if any(np.isnan(active_params[i])):
                nan_indices.append(i)
                min_f, max_f = find_extremes_f(f, active_params, i)
                lim_f1 = reference_params[i][1] - \
                    search_limit * reference_params[i][2]
                lim_f2 = reference_params[i][1] + \
                    search_limit * reference_params[i][2]
                min_lim_f = min(lim_f1, lim_f2)
                max_lim_f = max(lim_f1, lim_f2)
                min_f = max(min_f, min_lim_f)
                max_f = min(max_f, max_lim_f)
                min_f, max_f = min(min_f, max_f), max(min_f, max_f)
                nan_limits.append((min_f, max_f))
                min_ind = util.find_nearest_index(f, min_f)
                max_ind = util.find_nearest_index(f, max_f)
                reference_FWHM = reference_params[i][2]
                delta_f = max_f - min_f
                try:
                    base_zoom = np.abs(
                        int(np.ceil(np.log2(delta_f / reference_FWHM))))
                except BaseException:
                    base_zoom = 8
                if base_zoom > 8:
                    base_zoom = 8
                if base_zoom < 2:
                    base_zoom = 2
                known_zooms.append(base_zoom)
                known_zooms.append(base_zoom - 1)
                regions_one = np.array([[min_ind, max_ind]])
                regions = np.append(regions, regions_one, axis=0)
        input_regions = util.drop_region(regions)
        try:
            potential_regions = sw.split_peaks(
                model, f, v, input_regions, min(known_zooms), max(known_zooms))
        except BaseException:
            potential_regions = np.empty((0, 2))
        potential_params = fl.parameters_from_regions(
            f, v, potential_regions, noise_filter=noise_level)
        nan_params = np.empty((0, 4))
        for i in range(0, len(nan_indices)):
            new_params = np.array([reference_params[nan_indices[i]]])
            nan_params = np.append(nan_params, new_params, axis=0)
        chosen_params = match_params(nan_params, potential_params)
        checked_params = np.empty((0, 4))
        for i in range(0, len(nan_params)):
            if chosen_params[i][1] > nan_limits[i][0] and chosen_params[i][1] < nan_limits[i][1]:
                checked_params = np.append(
                    checked_params, [chosen_params[i]], axis=0)
            else:
                checked_params = np.append(
                    checked_params, [[np.nan, np.nan, np.nan, np.nan]], axis=0)
        updated_params = nan_merge(active_params, checked_params)
        return updated_params


def extend_params(params_2d, final_length):
    nan_array = np.empty((0, 4))
    for i in range(0, final_length - len(params_2d)):
        nan_array = np.append(
            nan_array, [[np.nan, np.nan, np.nan, np.nan]], axis=0)
    return np.append(params_2d, nan_array, axis=0)


def track_temperatures_learn(data_files, depth=None, live_display=False):
    sg.one_line_progress_meter_cancel('-key-')
    live = None
    all_peaks = []
    if depth is None:
        depth = len(data_files)
    for i in util._progressbar(range(0, depth), 'Processing: '):
        sg.one_line_progress_meter(
            'Overall Fitting Progress', i, depth, '-key-')
        parameters = automatic.quick_analyze(
            data_files[i].f, data_files[i].r, show=False, learn=True)
        if live_display:
            if live is not None:
                live.close_window()
            live = lf.Live_Instance(data_files[i].f, data_files[i].r)
            live.import_lorentzians(parameters)
            live.activate(loop=False)
        all_peaks.append(parameters)
    most_peaks_param = max(all_peaks, key=lambda a: len(a))
    most_peaks_length = len(most_peaks_param)
    for i in range(0, len(all_peaks)):
        all_peaks[i] = extend_params(all_peaks[i], most_peaks_length)
    all_peaks = np.array(all_peaks)
    sg.one_line_progress_meter_cancel('-key-')
    return all_peaks


def track_temperatures(
        data_files,
        initial_params=None,
        show=True,
        depth=None,
        learn=True,
        correction=True,
        start_index=0,
        print_params=False,
        search_limit=None,
        noise_limit=3,
        method='together',
        live_display=False,
        progress_window=True):
    """
    Tracks Lorentzians over a complete temperature sweep from some optional
    initial parameters.
    """
    # previous refers to the index before
    # reference refers to the last known usable values (could be index before or earlier)
    # active refers to the current index
    # search_limit doesn't currently do anything
    if progress_window:
        sg.one_line_progress_meter_cancel('-key-')
    live = None
    if initial_params is None:
        initial_params = np.empty((0, 4))
        initial_f = data_files[start_index].f
        initial_v = data_files[start_index].r
        initial_params = automatic.quick_analyze(
            initial_f, initial_v, show=show, learn=learn)
    if search_limit is None:
        search_limit = np.inf
    if method == 'separate':
        find_missing_peaks = find_missing_peaks_sep
    elif method == 'together':
        find_missing_peaks = find_missing_peaks_tog
    else:
        print('No valid peak refinding method given. Not doing any correction.')
        correction = False
    initial_params = util.param_sort(initial_params)
    all_peaks = np.array([initial_params])
    previous_regions = fl.regions_from_parameters(
        data_files[start_index].f, initial_params)
    if depth is None:
        depth = len(data_files) - start_index
    else:
        depth = min(depth, len(data_files) - start_index)
    if print_params:
        print('Initial Parameters: ')
        print(initial_params)
    for i in util._progressbar(
            range(start_index + 1, start_index + depth),
            'Processing: '):
        if progress_window:
            sg.one_line_progress_meter(
                'Overall Fitting Progress', i - start_index, depth, '-key-')
        f = data_files[i].f
        v = data_files[i].r
        reference_params, reference_counts = get_reference_params(all_peaks)
        noise_level = noise_limit * sw.extract_noise(v)
        active_params = fl.parameters_from_regions(
            f, v, previous_regions, noise_filter=noise_level)
        active_params = match_params(all_peaks[-1], active_params)
        active_params = util.param_sort(active_params)
        if correction:
            active_params = find_missing_peaks(
                f, v, active_params, reference_params, search_limit)
            active_params = match_params(reference_params, active_params)
            active_params = util.param_sort(active_params)
        active_params = util.param_sort(active_params)
        active_params = bind_params(
            reference_params, active_params, reference_counts)
        all_peaks = np.append(all_peaks, np.array([active_params]), axis=0)
        region_params, region_counts = get_reference_params(all_peaks)
        if live_display:
            if live is not None:
                live.close_window()
            live = lf.Live_Instance(f, v)
            live.import_lorentzians(region_params)
            live.activate(loop=False)
        if any(np.isnan(region_params.flatten())):
            region_params = util.remove_nans(region_params)
        previous_regions = fl.regions_from_parameters(
            data_files[i].f, region_params)
        previous_regions = bind_regions(f, previous_regions, region_params)
        previous_regions = util.drop_region(previous_regions)
    if live is not None:
        live.close_window()
    if progress_window:
        sg.one_line_progress_meter_cancel('-key-')
    return all_peaks


def bind_regions(f, regions, parameters):
    # regions and paarameters must be the same length
    bound_regions = np.empty((0, 2))
    for i in range(0, len(regions)):
        if i == 0:
            min_bound = 0
        else:
            min_bound = util.find_nearest_index(f, parameters[i - 1][1])
        if i == len(regions) - 1:
            max_bound = len(regions)
        else:
            max_bound = util.find_nearest_index(f, parameters[i + 1][1])
        min_ind = regions[i][0]
        max_ind = regions[i][1]
        min_ind = max(min_ind, min_bound)
        max_ind = min(max_ind, max_bound)
        bound_regions = np.append(bound_regions, [[min_ind, max_ind]], axis=0)
    return bound_regions


def bind_params(reference_params_2d, target_params_2d, reference_counts):
    bound_params = np.empty((0, 4))
    for i in range(0, len(reference_params_2d)):
        if params_bound_check(
                reference_params_2d[i],
                target_params_2d[i],
                reference_counts[i]):
            bound_params = np.append(
                bound_params, [target_params_2d[i]], axis=0)
        else:
            bound_params = np.append(
                bound_params, [[np.nan, np.nan, np.nan, np.nan]], axis=0)
    return bound_params


def params_bound_check(
        reference_params_1d,
        target_params_1d,
        reference_counts_1d):
    check_1 = bound_A_check(reference_params_1d,
                            target_params_1d, reference_counts_1d)
    check_2 = bound_f0_check(
        reference_params_1d, target_params_1d, reference_counts_1d)
    check_3 = bound_FWHM_check(
        reference_params_1d, target_params_1d, reference_counts_1d)
    check_4 = bound_phase_check(
        reference_params_1d, target_params_1d, reference_counts_1d)
    checks = [check_1, check_2, check_3, check_4]
    return all(checks)


def bound_A_check(reference_params_1d, target_params_1d, reference_counts_1d):
    return util.order_difference(
        reference_params_1d[0], target_params_1d[0]) < .35 * (1 + reference_counts_1d / 3)


def bound_f0_check(reference_params_1d, target_params_1d, reference_counts_1d):
    delta_f = np.abs(reference_params_1d[1] - target_params_1d[1])
    allowed_delta_f = np.abs(2 * reference_params_1d[2]) * reference_counts_1d
    return delta_f < allowed_delta_f


def bound_FWHM_check(
        refrence_params_1d,
        target_params_1d,
        reference_counts_1d):
    return util.order_difference(
        refrence_params_1d[2],
        target_params_1d[2]) < .5 * reference_counts_1d


def bound_phase_check(
        reference_params_1d,
        target_params_1d,
        reference_counts_1d):
    return np.abs(
        reference_params_1d[3] - target_params_1d[3]) < reference_counts_1d * np.pi / 2


def nan_merge(main_params, input_params):
    final_params = np.empty((0, 4))
    offset = 0
    for i in range(0, len(main_params)):
        if not any(np.isnan(main_params[i])):
            final_params = np.append(final_params, [main_params[i]], axis=0)
        else:
            final_params = np.append(
                final_params, [input_params[offset]], axis=0)
            offset += 1
    return final_params


def match_params(reference_params, target_params):
    """
    Matches the parameters between two parameter arrays.

    Parameters
    ----------
    reference_params : arr
        2D Lorentzian parameter array.
    target_params : arr
        2D Lorentzian parameter array.

    Returns
    -------
    updated_params : arr
        2D Lorentzian parameter array. Revised version of target_params.
    """
    ref_invert = False
    if len(target_params) > len(reference_params):
        reference_params, target_params = target_params, reference_params
        ref_invert = True
    reference_list = []
    target_list = []
    match_table = []
    # [ identifying index, parameters, matching index ]
    for i in range(0, len(reference_params)):
        reference_list.append([i, reference_params[i], None])
    for i in range(0, len(target_params)):
        target_list.append([i, target_params[i], None])
    for i in range(0, len(reference_params)):
        match_table.append([])
        for j in range(0, len(target_params)):
            ref_f0 = reference_params[i][1]
            tar_f0 = target_params[j][1]
            match_table[i].append(np.abs(ref_f0 - tar_f0))
    match_array = np.array(match_table)
    match_shape = match_array.shape
    best_vals = np.dstack(np.unravel_index(
        np.argsort(match_array.ravel()), match_shape))[0]
    num_matched = 0
    k_best = 0
    while num_matched < len(target_params):
        indices = best_vals[k_best]
        i = indices[0]
        j = indices[1]
        if update_match(reference_list, target_list, i, j):
            num_matched += 1
        k_best += 1
    if ref_invert:
        updated_params = get_updated_target(target_list, reference_list)
    else:
        updated_params = get_updated_target(reference_list, target_list)
    return updated_params


def update_match(reference_list, target_list, i, j):
    if reference_list[i][2] is None and target_list[j][2] is None:
        reference_list[i][2] = j
        target_list[j][2] = i
        return True
    else:
        return False


def get_updated_target(reference_list, target_list):
    updated_params = np.empty((0, 4))
    for i in range(0, len(reference_list)):
        j = reference_list[i][2]
        if j is not None:
            params_to_add = np.array([target_list[j][1]])
        else:
            params_to_add = np.array([[np.nan, np.nan, np.nan, np.nan]])
        updated_params = np.append(updated_params, params_to_add, axis=0)
    return updated_params


def update_params(param_list):
    """
    Given a list of parameters, will match them all up in order. See
    match_params for additional context.
    """
    for i in util._progressbar(range(1, len(param_list)), "Tracking: "):
        param_list[i] = match_params(param_list[i - 1], param_list[i])
    return param_list


def final_sorting(params_3d):
    """
    Sorts a 3D Lorentzian parameter array.

    Parameters
    ----------
    params_3d : arr
        3D Lorentzian parameter array.

    Returns
    -------
    final_params : arr
        3D Lorentzian parameter array.
    """
    final_params = np.array([params_3d[0]])
    for i in range(1, len(params_3d)):
        reference_params, counts = get_reference_params(params_3d[0:i])
        target_params = params_3d[i]
        matched_params = match_params(reference_params, target_params)
        combined_params = np.append(
            [reference_params], [target_params], axis=0)
        matched_params, counts = get_reference_params(combined_params)
        final_params = np.append(final_params, [matched_params], axis=0)
    return final_params
